from math import ceil

from addok.config import config
from addok.helpers import iter_pipe


def preprocess_query(s):
    return list(iter_pipe(s, config.QUERY_PROCESSORS + config.PROCESSORS))


def tokenize(helper):
    helper.tokens = preprocess_query(helper.query)
    if helper.tokens:
        helper.tokens[-1].is_last = True
        helper.last_token = helper.tokens[-1]
    helper.tokens.sort(key=lambda x: len(x), reverse=True)


def search_tokens(helper):
    for token in helper.tokens:
        token.search()


def set_should_match_threshold(helper):
    helper.should_match_threshold = ceil(2 / 3 * len(helper.tokens))


def select_tokens(helper):
    tokens = []
    for token in helper.tokens:
        if token.kind == "housenumber":
            helper.housenumbers.append(token)
            continue  # Remove from tokens (housenumbers are not indexed).
        elif token.is_common:
            helper.common.append(token)
        elif token.db_key:
            helper.meaningful.append(token)
        else:
            helper.not_found.append(token)
        tokens.append(token)
    helper.tokens = tokens
    helper.common.sort(key=lambda x: x.frequency)
    helper.meaningful.sort(key=lambda x: x.frequency)
    # Sanity limit.
    helper.common.extend(helper.meaningful[helper.MAX_MEANINGFUL :])
    helper.meaningful = helper.meaningful[: helper.MAX_MEANINGFUL]



import json
import logging
import logging.handlers
from pathlib import Path
import time

import falcon

from addok.config import config
from addok.core import reverse, search
from addok.db import DB
from addok.helpers.text import EntityTooLarge

notfound_logger = None
query_logger = None
slow_query_logger = None


def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    filename = Path(config.LOG_DIR).joinpath("{}.log".format(name))
    try:
        handler = logging.handlers.TimedRotatingFileHandler(
            str(filename), when="midnight"
        )
    except FileNotFoundError:
        print("Unable to write to {}".format(filename))
    else:
        logger.addHandler(handler)
    return logger


@config.on_load
def on_load():
    if config.LOG_NOT_FOUND:
        global notfound_logger
        notfound_logger = get_logger("notfound")

    if config.LOG_QUERIES:
        global query_logger
        query_logger = get_logger("queries")

    if config.SLOW_QUERIES:
        global slow_query_logger
        slow_query_logger = get_logger("slow_queries")


def log_notfound(query):
    if config.LOG_NOT_FOUND:
        notfound_logger.debug(query)


def log_query(query, results):
    if config.LOG_QUERIES:
        if results:
            result = str(results[0])
            score = str(round(results[0].score, 2))
        else:
            result = "-"
            score = "-"
        query_logger.debug("\t".join([query, result, score]))


def log_slow_query(query, results, timer):
    if config.SLOW_QUERIES:
        if results:
            result = str(results[0])
            score = str(round(results[0].score, 2))
            id_ = results[0].id
        else:
            result = "-"
            score = "-"
            id_ = "-"
        slow_query_logger.debug("\t".join([str(timer), query, id_, result, score]))


class CorsMiddleware:
    def process_response(self, req, resp, resource, req_succeeded):
        resp.set_header("Access-Control-Allow-Origin", "*")
        resp.set_header("Access-Control-Allow-Headers", "X-Requested-With")


class View:

    config = config

    def match_filters(self, req):
        filters = {}
        for name in config.FILTERS:
            req.get_param(name, store=filters)
        return filters

    def render(
        self, req, resp, results, query=None, filters=None, center=None, limit=None
    ):
        results = {
            "type": "FeatureCollection",
            "version": "draft",
            "features": [r.format() for r in results],
            "attribution": config.ATTRIBUTION,
            "licence": config.LICENCE,
        }
        if query:
            results["query"] = query
        if filters:
            results["filters"] = filters
        if center:
            results["center"] = center
        if limit:
            results["limit"] = limit
        self.json(req, resp, results)

    to_geojson = render  # retrocompat.

    def json(self, req, resp, content):
        resp.text = json.dumps(content)
        resp.content_type = "application/json; charset=utf-8"

    def parse_float(sel, req, *keys):
        try:
            for key in keys:
                val = req.get_param(key)
                if val is not None:
                    return float(val)
        except (ValueError, TypeError):
            raise falcon.HTTPInvalidParam("invalid value", key)
        return None

    def parse_lon_lat(self, req):
        lat = self.parse_float(req, "lat", "latitude")
        lon = self.parse_float(req, "lon", "lng", "long", "longitude")

        if lon and (lon > 180 or lon < -180):
            raise falcon.HTTPInvalidParam("out of range", "lon")
        elif lat and (lat > 90 or lat < -90):
            raise falcon.HTTPInvalidParam("out of range", "lat")
        return lon, lat


class Search(View):
    def on_get(self, req, resp, **kwargs):
        query = req.get_param("q")
        if not query:
            raise falcon.HTTPMissingParam("q")
        limit = req.get_param_as_int("limit") or 5  # use config
        if limit < 1 or limit > 100:
            raise falcon.HTTPInvalidParam("out of range (1..100)", "limit")
        autocomplete = req.get_param_as_bool("autocomplete")
        if autocomplete is None:
            # Default is True.
            # https://github.com/falconry/falcon/pull/493#discussion_r44376219
            autocomplete = True
        lon, lat = self.parse_lon_lat(req)
        center = None
        if lon and lat:
            center = (lon, lat)
        filters = self.match_filters(req)
        timer = time.perf_counter()
        try:
            results = search(
                query,
                limit=limit,
                autocomplete=autocomplete,
                lat=lat,
                lon=lon,
                **filters
            )
        except EntityTooLarge as e:
            raise falcon.HTTPContentTooLarge(title=str(e))
        timer = int((time.perf_counter() - timer) * 1000)
        if not results:
            log_notfound(query)
        log_query(query, results)
        if config.SLOW_QUERIES and timer > config.SLOW_QUERIES:
            log_slow_query(query, results, timer)
        self.render(
            req, resp, results, query=query, filters=filters, center=center, limit=limit
        )


class Reverse(View):
    def on_get(self, req, resp, **kwargs):
        lon, lat = self.parse_lon_lat(req)
        if lon is None:
            raise falcon.HTTPMissingParam("lon")
        if lat is None:
            raise falcon.HTTPMissingParam("lat")
        limit = req.get_param_as_int("limit") or 1
        filters = self.match_filters(req)
        results = reverse(lat=lat, lon=lon, limit=limit, **filters)
        self.render(req, resp, results, filters=filters, limit=limit)


class Health(View):
    def on_get(self, req, resp):
        return self.json(
            req,
            resp,
            {"status": "HEALTHY", "redis_version": DB.info().get("redis_version")},
        )


def register_http_endpoint(api):
    api.add_route("/search", Search())
    api.add_route("/reverse", Reverse())
    api.add_route("/health", Health())


def register_command(subparsers):
    parser = subparsers.add_parser("serve", help="Run debug server")
    parser.set_defaults(func=run)
    parser.add_argument(
        "--host", default="127.0.0.1", help="Host to expose the demo serve on"
    )
    parser.add_argument(
        "--port", default="7878", help="Port to expose the demo server on"
    )


def run(args):
    # Do not import at load time for preventing config import loop.
    from .wsgi import simple

    simple(args)



import os
from pathlib import Path

REDIS = {
    "host": os.environ.get("REDIS_HOST") or "localhost",
    "port": os.environ.get("REDIS_PORT") or 6379,
    "unix_socket_path": os.environ.get("REDIS_SOCKET"),
    "indexes": {
        "db": os.environ.get("REDIS_DB_INDEXES") or 0,
    },
    "documents": {
        "db": os.environ.get("REDIS_DB_DOCUMENTS") or 1,
    },
}

# Min/max number of results to be retrieved from db and scored.
BUCKET_MIN = 10
BUCKET_MAX = 100

# Maximum number of values allowed in a multi-value filter (e.g., type=street+city+locality).
# This limit prevents performance issues and potential abuse.
MAX_FILTER_VALUES = 10

# Above this threshold, terms are considered commons.
COMMON_THRESHOLD = 10000

# Above this threshold, we avoid intersecting sets.
INTERSECT_LIMIT = 100000

# Min score considered matching the query.
MATCH_THRESHOLD = 0.9

# Do not consider result if final score is below this threshold.
MIN_SCORE = 0.1

QUERY_MAX_LENGTH = 200

GEOHASH_PRECISION = 7

MIN_EDGE_NGRAMS = 3
MAX_EDGE_NGRAMS = 20

SYNONYMS_PATHS = []

# Pipeline stream to be used.
PROCESSORS_PYPATHS = [  # Rename in TOKEN_PROCESSORS / STRING_PROCESSORS?
    "addok.helpers.text.tokenize",
    "addok.helpers.text.normalize",
    "addok.helpers.text.flag_housenumber",
    "addok.helpers.text.synonymize",
]
QUERY_PROCESSORS_PYPATHS = [
    "addok.helpers.text.check_query_length",
]
# Remove SEARCH_PREFIXES.
SEARCH_PREPROCESSORS_PYPATHS = [
    "addok.helpers.search.tokenize",
    "addok.helpers.search.search_tokens",
    "addok.helpers.search.select_tokens",
    "addok.helpers.search.set_should_match_threshold",
]
BATCH_PROCESSORS_PYPATHS = [
    "addok.batch.to_json",
    "addok.helpers.index.prepare_housenumbers",
    "addok.ds.store_documents",
    "addok.helpers.index.index_documents",
]
BATCH_FILE_LOADER_PYPATH = "addok.helpers.load_file"
BATCH_CHUNK_SIZE = 1000
# During imports, workers are consuming RAM;
# let one process free for Redis by default.
BATCH_WORKERS = max(os.cpu_count() - 1, 1)
RESULTS_COLLECTORS_PYPATHS = [
    "addok.autocomplete.only_commons_but_geohash_try_autocomplete_collector",
    "addok.helpers.collectors.no_tokens_but_housenumbers_and_geohash",
    "addok.helpers.collectors.no_available_tokens_abort",
    "addok.helpers.collectors.only_commons",
    "addok.autocomplete.no_meaningful_but_common_try_autocomplete_collector",
    "addok.autocomplete.only_commons_try_autocomplete_collector",
    "addok.helpers.collectors.bucket_with_meaningful",
    "addok.helpers.collectors.reduce_with_other_commons",
    "addok.helpers.collectors.ensure_geohash_results_are_included_if_center_is_given",  # noqa
    "addok.autocomplete.autocomplete_meaningful_collector",
    "addok.fuzzy.fuzzy_collector",
    "addok.helpers.collectors.extend_results_extrapoling_relations",
    "addok.helpers.collectors.extend_results_reducing_tokens",
]
SEARCH_RESULT_PROCESSORS_PYPATHS = [
    "addok.helpers.results.match_housenumber",
    "addok.helpers.results.make_labels",
    "addok.helpers.results.score_by_importance",
    "addok.helpers.results.score_by_autocomplete_distance",
    "addok.helpers.results.score_by_ngram_distance",
    "addok.helpers.results.score_by_geo_distance",
    "addok.helpers.results.adjust_scores",
]
REVERSE_RESULT_PROCESSORS_PYPATHS = [
    "addok.helpers.results.load_closer",
    "addok.helpers.results.make_labels",
    "addok.helpers.results.score_by_geo_distance",
]
RESULTS_FORMATTERS_PYPATHS = [
    "addok.helpers.formatters.geojson",
]
INDEXERS_PYPATHS = [
    "addok.helpers.index.HousenumbersIndexer",
    "addok.helpers.index.FieldsIndexer",
    # Pairs indexer must be after `FieldsIndexer`.
    "addok.pairs.PairsIndexer",
    # Edge ngram indexer must be after `FieldsIndexer`.
    "addok.autocomplete.EdgeNgramIndexer",
    "addok.helpers.index.FiltersIndexer",
    "addok.helpers.index.GeohashIndexer",
]
# Any object like instance having `loads` and `dumps` methods.
DOCUMENT_SERIALIZER_PYPATH = "addok.helpers.serializers.ZlibSerializer"

DOCUMENT_STORE_PYPATH = "addok.ds.RedisStore"

# Fields to be indexed
# If you want a housenumbers field but need to name it differently, just add
# type="housenumbers" to your field.
FIELDS = [
    {"key": "name", "boost": 4, "null": False},
    {"key": "street"},
    {
        "key": "postcode",
        "boost": lambda doc: 1.2 if doc.get("type") == "municipality" else 1,
    },
    {"key": "city"},
    {"key": "housenumbers"},
    {"key": "context"},
]
ID_FIELD = "_id"

# Sometimes you only want to add some fields keeping the default ones.
EXTRA_FIELDS = []

# Weight of a document own importance:
IMPORTANCE_WEIGHT = 0.1

# Geographical distance importance on final score
GEO_DISTANCE_WEIGHT = 0.1

# Default score for the relation token => document
DEFAULT_BOOST = 1.0

# Data attribution
# Can also be an object {source: attribution}
ATTRIBUTION = "BANO"

# Data licence
# Can also be an object {source: licence}
LICENCE = "ODbL"

# Available filters (remember that every filter means bigger index)
FILTERS = ["type", "postcode"]

LOG_DIR = os.environ.get("ADDOK_LOG_DIR", Path(__file__).parent.parent.parent)

LOG_QUERIES = False
LOG_NOT_FOUND = False
SLOW_QUERIES = False  # False or time in ms to consider query as slow

INDEX_EDGE_NGRAMS = True

# surrounding letters on a standard keyboard (default french azerty)
FUZZY_KEY_MAP = {
    "a": "ezqop",
    "z": "aqse",
    "e": "azsdryu",
    "r": "edft",
    "t": "rfgy",
    "y": "teghu",
    "u": "yehji",
    "i": "ujko",
    "o": "iaklp",
    "p": "oalm",
    "q": "azsw",
    "s": "qzedxw",
    "d": "serfcx",
    "f": "drtgvc",
    "g": "ftyhbv",
    "h": "gyujnb",
    "j": "huikn",
    "k": "jil",
    "l": "kom",
    "m": "lpu",
    "w": "qsx",
    "x": "wsdc",
    "c": "xdfvio",
    "v": "cfgb",
    "b": "vghn",
    "n": "bhj",
}



"""Tests for multiprocessing helpers."""
import os
import sys
from unittest.mock import MagicMock, patch

import pytest

from addok.config import config
from addok.helpers import _get_config_overrides, _worker_init


def test_get_config_overrides_detects_changed_values():
    """Test that _get_config_overrides detects configuration changes."""
    # Modify a config value
    original_value = config.MIN_EDGE_NGRAMS
    config.MIN_EDGE_NGRAMS = 99

    try:
        overrides = _get_config_overrides()

        # Should detect the change
        assert 'MIN_EDGE_NGRAMS' in overrides
        assert overrides['MIN_EDGE_NGRAMS'] == 99
    finally:
        # Restore original value
        config.MIN_EDGE_NGRAMS = original_value


def test_get_config_overrides_skips_unchanged_values():
    """Test that unchanged values are not included in overrides."""
    overrides = _get_config_overrides()

    # BUCKET_MIN should not be in overrides if it hasn't changed from default
    # (unless it was changed in test config)
    from addok.config import default as default_config

    if config.BUCKET_MIN == default_config.BUCKET_MIN:
        assert 'BUCKET_MIN' not in overrides


def test_get_config_overrides_skips_functions():
    """Test that functions are filtered out from overrides."""
    # Set a function as a config value
    def dummy_func():
        pass

    config.TEST_FUNCTION = dummy_func

    # Functions should be filtered out
    overrides = _get_config_overrides()
    assert 'TEST_FUNCTION' not in overrides

    # Clean up
    del config['TEST_FUNCTION']


def test_get_config_overrides_skips_classes():
    """Test that classes are filtered out from overrides."""
    class DummyClass:
        pass

    config.TEST_CLASS = DummyClass

    # Classes should be filtered out
    overrides = _get_config_overrides()
    assert 'TEST_CLASS' not in overrides

    # Clean up
    del config['TEST_CLASS']


def test_get_config_overrides_includes_serializable_values():
    """Test that serializable values are included."""
    config.TEST_STRING = "test_value"
    config.TEST_INT = 42
    config.TEST_BOOL = True
    config.TEST_LIST = [1, 2, 3]
    config.TEST_DICT = {"key": "value"}

    try:
        overrides = _get_config_overrides()

        # All these should be included (if different from defaults)
        from addok.config import default as default_config

        if not hasattr(default_config, 'TEST_STRING'):
            assert overrides.get('TEST_STRING') == "test_value"
        if not hasattr(default_config, 'TEST_INT'):
            assert overrides.get('TEST_INT') == 42
        if not hasattr(default_config, 'TEST_BOOL'):
            assert overrides.get('TEST_BOOL') is True
        if not hasattr(default_config, 'TEST_LIST'):
            assert overrides.get('TEST_LIST') == [1, 2, 3]
        if not hasattr(default_config, 'TEST_DICT'):
            assert overrides.get('TEST_DICT') == {"key": "value"}
    finally:
        # Clean up
        for key in ['TEST_STRING', 'TEST_INT', 'TEST_BOOL', 'TEST_LIST', 'TEST_DICT']:
            if key in config:
                del config[key]


@patch('addok.helpers.os.environ', {})
def test_worker_init_sets_environment_variables():
    """Test that _worker_init sets environment variables."""
    env_vars = {'TEST_VAR': 'test_value'}
    redis_params = {
        'indexes': {'host': 'localhost', 'port': 6379, 'db': 0},
        'documents': {'host': 'localhost', 'port': 6379, 'db': 1},
        'use_redis_documents': False
    }

    with patch('addok.db.DB.connect') as mock_connect:
        _worker_init(redis_params, config_env_vars=env_vars)

        assert os.environ.get('TEST_VAR') == 'test_value'
        mock_connect.assert_called_once()


def test_worker_init_applies_config_overrides():
    """Test that _worker_init applies configuration overrides."""
    redis_params = {
        'indexes': {'host': 'localhost', 'port': 6379, 'db': 0},
        'documents': {'host': 'localhost', 'port': 6379, 'db': 1},
        'use_redis_documents': False
    }
    overrides = {'MIN_EDGE_NGRAMS': 99}

    original_value = config.MIN_EDGE_NGRAMS

    try:
        with patch('addok.db.DB.connect'):
            _worker_init(redis_params, config_overrides=overrides)

            # Config should be updated
            assert config.MIN_EDGE_NGRAMS == 99
    finally:
        # Restore
        config.MIN_EDGE_NGRAMS = original_value


def test_worker_init_connects_to_redis():
    """Test that _worker_init connects to Redis."""
    redis_params = {
        'indexes': {'host': 'localhost', 'port': 6379, 'db': 14},
        'documents': {'host': 'localhost', 'port': 6379, 'db': 15},
        'use_redis_documents': True
    }

    with patch('addok.db.DB.connect') as mock_db_connect:
        with patch('addok.ds._DB.connect') as mock_ds_connect:
            _worker_init(redis_params)

            # Should connect to both databases
            mock_db_connect.assert_called_once_with(
                host='localhost', port=6379, db=14
            )
            mock_ds_connect.assert_called_once_with(
                host='localhost', port=6379, db=15
            )


def test_worker_init_skips_document_store_if_not_redis():
    """Test that document store connection is skipped when not using Redis."""
    redis_params = {
        'indexes': {'host': 'localhost', 'port': 6379, 'db': 14},
        'documents': {'host': 'localhost', 'port': 6379, 'db': 15},
        'use_redis_documents': False
    }

    with patch('addok.db.DB.connect') as mock_db_connect:
        with patch('addok.ds._DB.connect') as mock_ds_connect:
            _worker_init(redis_params)

            # Should only connect to indexes DB
            mock_db_connect.assert_called_once()
            mock_ds_connect.assert_not_called()


@pytest.mark.skipif(sys.platform != "darwin", reason="macOS-specific test")
def test_parallelize_uses_spawn_on_macos():
    """Test that parallelize uses spawn context on macOS."""
    from addok.helpers import parallelize

    def dummy_func(*items):
        return items

    # This is a smoke test - just ensure it doesn't crash
    # We can't easily test the actual multiprocessing without mocking
    with patch('addok.helpers.ChunkedPool') as mock_pool:
        mock_pool_instance = MagicMock()
        mock_pool.return_value.__enter__ = MagicMock(return_value=mock_pool_instance)
        mock_pool.return_value.__exit__ = MagicMock(return_value=False)
        mock_pool_instance.imap_unordered.return_value = []

        parallelize(dummy_func, [], chunk_size=10)

        # Verify spawn context was requested on macOS
        from multiprocessing import get_context
        mock_pool.assert_called_once()
        call_kwargs = mock_pool.call_args[1]
        assert 'context' in call_kwargs
        # Context should be spawn on macOS
        assert call_kwargs['context']._name == 'spawn'


@pytest.mark.skipif(sys.platform == "darwin", reason="Linux-specific test")
def test_parallelize_uses_fork_on_linux():
    """Test that parallelize uses fork context on Linux."""
    from addok.helpers import parallelize

    def dummy_func(*items):
        return items

    with patch('addok.helpers.ChunkedPool') as mock_pool:
        mock_pool_instance = MagicMock()
        mock_pool.return_value.__enter__ = MagicMock(return_value=mock_pool_instance)
        mock_pool.return_value.__exit__ = MagicMock(return_value=False)
        mock_pool_instance.imap_unordered.return_value = []

        parallelize(dummy_func, [], chunk_size=10)

        # Verify fork context was requested on Linux
        mock_pool.assert_called_once()
        call_kwargs = mock_pool.call_args[1]
        assert 'context' in call_kwargs
        # Context should be fork on Linux
        assert call_kwargs['context']._name == 'fork'



from addok.helpers.collectors import _extract_manytomany_relations
from addok.helpers.text import Token


def test_extract_manytomany_relations(factory, config):
    config.COMMON_THRESHOLD = 2
    factory(name="rue de Paris", city="Fecamp")
    factory(name="rue de la porte")
    factory(name="rue de dieppe", housenumbers={"506": {"lat": 1, "lon": 2}})
    tokens = [Token(s) for s in "rue de paris porte 506 fecamp".split()]
    groups = _extract_manytomany_relations(tokens)
    assert groups == [{Token("fecamp"), Token("paris")}]


def test_extract_manytomany_relations_2(factory, config):
    config.COMMON_THRESHOLD = 2
    factory(
        name="rue de falaise",
        city="dieppe",
        postcode="76370",
        housenumbers={"1": {"lat": 1, "lon": 2}},
    )
    factory(
        name="chemin du semaphore",
        city="dieppe",
        postcode="76370",
        housenumbers={"1": {"lat": 1, "lon": 2}},
    )
    factory(name="chemin de neuville")
    factory(name="chemin de la tour", housenumbers={"1": {"lat": 1, "lon": 2}})
    tokens = [
        Token(s)
        for s in "1 chemin de la falaise le semaphore neuville les 76370 "
        "dieppe".split()
    ]
    groups = _extract_manytomany_relations(tokens)
    assert len(groups) == 2
    assert {Token("dieppe"), Token("falaise"), Token("76370")} in groups
    assert {Token("dieppe"), Token("76370"), Token("semaphore")} in groups


def test_extract_manytomany_relations_3(factory, config):
    config.COMMON_THRESHOLD = 2
    latlon = {"lat": 1, "lon": 2}
    factory(
        name="Rue Maréchal de Lattre de Tassigny",
        city="Mont-Saint-Aignan",
        postcode="76130",
        housenumbers={"45": latlon, "3": latlon},
    )
    factory(
        name="rue du port",
        city="Saint-Denis",
        postcode="76370",
        housenumbers={"45": latlon, "3": latlon},
    )
    factory(name="rue à l'eau", city="Saint-Pierre-de-Rouergue")
    factory(name="rue de Saint-Jean", housenumbers={"45": latlon, "3": latlon})
    tokens = [
        Token(s)
        for s in "45 rue de lattre de tassign pleiade a 3 porte 76130 mont "
        "saint aignan".split()
    ]
    groups = _extract_manytomany_relations(tokens)
    assert groups == [
        {Token("lattre"), Token("aignan"), Token("76130"), Token("mont")},
    ]



import redis
from hashids import Hashids

from addok.config import config


hashids = Hashids()


class RedisProxy:
    instance = None
    Error = redis.RedisError

    def connect(self, *args, **kwargs):
        self.instance = redis.Redis(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def next_id(self):
        next_id = self.incr("_id_sequence")
        return hashids.encode(next_id)


DB = RedisProxy()


def _extract_redis_config(config_section):
    """Extract Redis connection parameters from a config section.

    Args:
        config_section: Dict with Redis configuration

    Returns:
        Dict with connection parameters
    """
    return {
        "host": config_section.get("host"),
        "port": config_section.get("port"),
        "db": config_section.get("db"),
        "password": config_section.get("password"),
        "unix_socket_path": config_section.get("unix_socket_path"),
    }


def get_redis_params():
    """Extract Redis connection parameters from config for multiprocessing workers.

    Returns:
        Dict with separate connection parameters for indexes and documents databases.
    """
    from addok import ds

    # Get indexes (main DB) parameters
    indexes_params = config.REDIS.copy()
    indexes_params.update(config.REDIS.get("indexes", {}))

    # Get documents DB parameters
    documents_params = config.REDIS.copy()
    documents_params.update(config.REDIS.get("documents", {}))

    return {
        "indexes": _extract_redis_config(indexes_params),
        "documents": _extract_redis_config(documents_params),
        "use_redis_documents": config.DOCUMENT_STORE == ds.RedisStore,
    }


@config.on_load
def connect():
    params = get_redis_params()
    DB.connect(**params['indexes'])



from addok.core import reverse


def test_reverse_return_closer_point(factory):
    factory(lat=78.23, lon=-15.23)
    good = factory(lat=48.234545, lon=5.235445)
    assert reverse(lat=48.234545, lon=5.235445)[0].id == good["id"]


def test_reverse_return_housenumber(factory):
    factory(housenumbers={"24": {"lat": 48.234545, "lon": 5.235445, "key": "value"}})
    results = reverse(lat=48.234545, lon=5.235445)
    assert results[0].housenumber == "24"
    assert results[0].type == "housenumber"
    assert results[0].key == "value"
    assert not results[0].raw


def test_reverse_can_be_limited(factory):
    factory(lat=48.234545, lon=5.235445)
    factory(lat=48.234546, lon=5.235446)
    results = reverse(lat=48.234545, lon=5.235445)
    assert len(results) == 1
    results = reverse(lat=48.234545, lon=5.235445, limit=2)
    assert len(results) == 2


def test_reverse_can_be_filtered(factory):
    factory(lat=48.234545, lon=5.235445, type="street")
    factory(lat=48.234546, lon=5.235446, type="city")
    results = reverse(lat=48.234545, lon=5.235445, type="city")
    assert len(results) == 1
    assert results[0].type == "city"


def test_reverse_supports_multi_value_filter(factory):
    """Test that reverse geocoding supports multi-value filters with OR logic"""
    street = factory(lat=48.234545, lon=5.235445, type="street")
    city = factory(lat=48.234546, lon=5.235446, type="city")
    locality = factory(lat=48.234547, lon=5.235447, type="locality")
    results = reverse(lat=48.234545, lon=5.235445, type="street+city", limit=10)
    assert len(results) == 2
    ids = {r.id for r in results}
    assert street["id"] in ids
    assert city["id"] in ids
    assert locality["id"] not in ids


def test_reverse_multi_filter_combination(factory):
    """Test that reverse geocoding combines multiple filters with AND logic"""
    match1 = factory(lat=48.234545, lon=5.235445, type="street", postcode="75001")
    match2 = factory(lat=48.234546, lon=5.235446, type="city", postcode="75001")
    no_match = factory(lat=48.234547, lon=5.235447, type="street", postcode="75002")
    results = reverse(
        lat=48.234545, lon=5.235445, type="street+city", postcode="75001", limit=10
    )
    assert len(results) == 2
    ids = {r.id for r in results}
    assert match1["id"] in ids
    assert match2["id"] in ids
    assert no_match["id"] not in ids


def test_reverse_should_not_return_housenumber_if_filtered(factory):
    factory(
        lat=48.234544,
        lon=5.235444,
        housenumbers={"24": {"lat": 48.234545, "lon": 5.235445}},
    )
    results = reverse(lat=48.234545, lon=5.235445, type="street")
    assert results[0].type == "street"
    results = reverse(lat=48.234545, lon=5.235445)
    assert results[0].type == "housenumber"
    results = reverse(lat=48.234545, lon=5.235445, type="housenumber")
    assert results[0].type == "housenumber"


def test_reverse_should_enforce_housenumber_if_filtered(factory):
    factory(
        lat=48.234544,
        lon=5.235444,
        housenumbers={"24": {"lat": 48.234545, "lon": 5.235445}},
    )
    results = reverse(lat=48.234544, lon=5.235444, type="housenumber")
    assert results[0].type == "housenumber"



"""
Exposes a pytest plugin so addok plugins do not need to recode all the test
setup logic.
"""

import json
import os
from pathlib import Path
import uuid

import pytest

# Do not import files from the top of the module, otherwise they will
# not be taken into account by the coverage.


def pytest_configure():
    from addok.config import config as addok_config

    addok_config.__class__.TESTING = True
    # Force test config.
    from addok import config as config_module

    os.environ["ADDOK_CONFIG_MODULE"] = str(
        Path(config_module.__file__).parent / "test.py"
    )
    import logging

    logging.basicConfig(level=logging.DEBUG)
    addok_config.REDIS["indexes"]["db"] = 14
    addok_config.REDIS["documents"]["db"] = 15
    addok_config.load()


def pytest_runtest_setup(item):
    from addok import db, ds
    from addok.config import config as addok_config

    assert db.DB.connection_pool.connection_kwargs["db"] == 14
    if addok_config.DOCUMENT_STORE == ds.RedisStore:
        assert ds._DB.connection_pool.connection_kwargs["db"] == 15


def pytest_runtest_teardown(item, nextitem):
    from addok import db, ds
    from addok.config import config as addok_config

    assert db.DB.connection_pool.connection_kwargs["db"] == 14
    db.DB.flushdb()
    if addok_config.DOCUMENT_STORE == ds.RedisStore:
        assert ds._DB.connection_pool.connection_kwargs["db"] == 15
        ds._DB.flushdb()


def pytest_addoption(parser):
    group = parser.getgroup("general")
    group._addoption(
        "--shell",
        action="store_true",
        dest="addokshell",
        default=False,
        help="start addok interactive shell on error.",
    )


def pytest_exception_interact(node, call, report):
    if node.config.getvalue("addokshell"):
        from addok.shell import invoke

        invoke()


class DummyDoc(dict):
    def __init__(self, *args, **kwargs):
        skip_index = kwargs.pop("skip_index", False)
        super().__init__(*args, **kwargs)
        if not skip_index:
            self.index()

    def update(self, **kwargs):
        super().update(kwargs)
        self.index()

    def index(self):
        from addok.batch import process_documents

        process_documents(json.dumps(self.copy()))


@pytest.fixture
def factory(request):
    from addok import db

    def _(**kwargs):
        default = {
            "id": uuid.uuid4().hex,
            "_id": db.DB.next_id(),
            "type": "street",
            "name": "ellington",
            "importance": 0.0,
            "lat": "48.3254",
            "lon": "2.256",
        }
        default.update(kwargs)
        doc = DummyDoc(**default)
        return doc

    return _


@pytest.fixture
def street(factory):
    return factory()


@pytest.fixture
def city(factory):
    return factory(type="city")


@pytest.fixture
def housenumber(factory):
    return factory(housenumbers={"11": {"lat": "48.3254", "lon": "2.256"}})


@pytest.fixture
def app():
    # Do not import before redis config has been
    # patched.
    from addok.http.wsgi import application

    return application


@pytest.fixture
def client(app):
    """Provide a Falcon test client accepting dict query strings."""
    from falcon import testing
    from urllib.parse import urlencode

    class _Client(testing.TestClient):
        def get(self, path, **kwargs):
            qs = kwargs.get("query_string")
            if isinstance(qs, dict):
                # simulate the behavior of pytest-falcon which accepted dicts
                kwargs["query_string"] = urlencode(qs, doseq=True)
            return super().get(path, **kwargs)

    return _Client(app)


class MonkeyPatchWrapper(object):
    def __init__(self, monkeypatch, wrapped_object):
        super().__setattr__("monkeypatch", monkeypatch)
        super().__setattr__("wrapped_object", wrapped_object)

    def __getattr__(self, attr):
        return getattr(self.wrapped_object, attr)

    def __setattr__(self, attr, value):
        self.monkeypatch.setattr(self.wrapped_object, attr, value, raising=False)

    def __delattr__(self, attr):
        self.monkeypatch.delattr(self.wrapped_object, attr)


@pytest.fixture()
def config(request, monkeypatch):
    from addok.config import config as addok_config

    return MonkeyPatchWrapper(monkeypatch, addok_config)



import atexit
import cmd
import json
import logging
import re
import time
import os

import geohash

from . import hooks
from .config import config
from .core import Result, Search, compute_geohash_key, reverse
from .db import DB
from .ds import get_document
from .helpers import (
    blue,
    cyan,
    green,
    haversine_distance,
    keys,
    km_to_score,
    magenta,
    red,
    scripts,
    white,
    yellow,
)
from .helpers.index import token_frequency
from .helpers.search import preprocess_query
from .helpers.text import compare_str

try:
    import gnureadline as readline  # For OSX.
except ImportError:
    import readline  # Normal way.


class Cmd(cmd.Cmd):

    intro = (
        white("\nWelcome to the Addok shell o/\n")
        + magenta("Type HELP or ? to list commands.\n")
        + magenta("Type QUIT or ctrl-C or ctrl-D to quit.\n")
    )
    prompt = "> "
    HISTORY_FILE = ".addok_shell_history"

    def __init__(self):
        self._init_history_file()
        super().__init__()
        self.do_HELP = self.do_help

    def _init_history_file(self):
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(self.history_file)
            except FileNotFoundError:
                pass
            atexit.register(self.save_history)

    def save_history(self):
        try:
            readline.write_history_file(self.history_file)
        except FileNotFoundError:
            print(
                red("Unable to write history file to " "{}.".format(self.history_file))
            )

    @property
    def history_file(self):
        return os.path.join(os.path.expanduser("~"), self.HISTORY_FILE)

    def error(self, message):
        print(red(message))

    def default(self, line):
        if line == "EOF":
            return self.quit()
        return self.do_SEARCH(line)

    def quit(self):
        print(red("Bye!"))
        return True

    def loop(self, *args):
        try:
            self.cmdloop()
        except KeyboardInterrupt:
            return self.quit()

    def completenames(self, text, *ignored):
        return super().completenames(text.upper(), *ignored)

    def get_names(self):
        special = ["do_help", "do_QUIT"]
        return [n for n in super().get_names() if n not in special]

    def postcmd(self, stop, line):
        if line != "EOF":
            print(yellow("-" * 80))
        return super().postcmd(stop, line)

    @classmethod
    def register_command(cls, command, name=None, doc=None):
        if name is None:
            name = command.__name__
        if doc is None:
            doc = command.__doc__
        if name.startswith("do_"):
            name = name[3:]
        name = "do_" + name.upper()
        setattr(cls, name, command)
        attr = getattr(cls, name)
        attr.__doc__ = doc

    @classmethod
    def register_commands(cls, *commands, **named):
        for command in commands:
            cls.register_command(command)
        for name, command in named.items():
            cls.register_command(command, name)

    def do_help(self, command):
        """Display this help message."""
        if command:
            doc = getattr(self, "do_" + command).__doc__
            print(cyan(doc.replace(" " * 8, "")))
        else:
            print(magenta("Available commands:"))
            print(magenta('Type "HELP <command>" to get more info.'))
            names = self.get_names()
            names.sort()
            for name in names:
                if name[:3] != "do_":
                    continue
                doc = getattr(self, name).__doc__
                doc = doc.split("\n")[0]
                print(
                    "{} {}".format(
                        yellow(name[3:]),
                        cyan(doc.replace(" " * 8, " ").replace("\n", "")),
                    )
                )

    @staticmethod
    def _match_option(key, string):
        matchs = re.findall("{}[= ][^ ]*".format(key), string)
        option = None
        if matchs:
            option = matchs[0]
            string = string.replace(option, "")
            option = option.replace(key, "")
        return string.strip(), option.strip(" =") if option else option

    def _search(self, query, verbose=False, bucket=False, count=1):
        limit = 10
        autocomplete = True
        lat = None
        lon = None
        filters = {}
        if "AUTOCOMPLETE" in query:
            query, autocomplete = self._match_option("AUTOCOMPLETE", query)
            autocomplete = bool(int(autocomplete))
        if "LIMIT" in query:
            query, limit = self._match_option("LIMIT", query)
            limit = int(limit)
        if "CENTER" in query:
            query, center = self._match_option("CENTER", query)
            lat, lon = center.split()
            lat = float(lat)
            lon = float(lon)
        for name in config.FILTERS:
            name = name.upper()
            if name in query:
                query, value = self._match_option(name, query)
                if value:
                    filters[name.lower()] = value.strip()
        helper = Search(limit=limit, verbose=verbose, autocomplete=autocomplete)
        start = time.time()
        for i in range(0, count):
            results = helper(query, lat=lat, lon=lon, **filters)
        if bucket:  # Means we want all the bucket
            results = helper._sorted_bucket
        duration = round((time.time() - start) * 1000 / count, 1)
        if verbose:
            helper.report()

        def format_scores(result):
            if verbose or bucket:
                return ", ".join(
                    "{}: {}/{}".format(k, round(v[0], 4), v[1])
                    for k, v in result._scores.items()
                )
            else:
                return result.score

        for result in results:
            print(
                "{} ({} | {})".format(
                    white(result), blue(result._id), blue(format_scores(result))
                )
            )
        formatter = red if duration > 50 else green
        print(
            "{} — {} — {}".format(
                formatter("{} ms".format(duration)),
                magenta("{} run(s)".format(count)),
                cyan("{} results".format(len(results))),
            )
        )

    def do_QUIT(self, *args):
        """Quit this shell. Also ctrl-C or Ctrl-D."""
        return self.quit()

    def do_SEARCH(self, query):
        """Issue a search (default command, can be omitted):
        SEARCH rue des Lilas [CENTER lat lon] [LIMIT 10] [AUTOCOMPLETE 0] [FILTER VALUE…]"""  # noqa
        self._search(query)

    def do_EXPLAIN(self, query):
        """Issue a search with debug info:
        EXPLAIN rue des Lilas"""
        self._search(query, verbose=True)

    def do_BUCKET(self, query):
        """Issue a search and return all the collected bucket, not only up to
        limit elements:
        BUCKET rue des Lilas"""
        self._search(query, bucket=True)

    def do_BENCH(self, query):
        """Run a search many times to benchmark it.
        BENCH [100] rue des Lilas"""
        try:
            count = int(re.match(r"^(\d+).*", query).group(1))
        except AttributeError:
            count = 100
        self._search(query, count=count)

    def do_INTERSECT(self, words):
        """Do a raw intersect between tokens (default limit 100).
        INTERSECT rue des lilas [LIMIT 100]"""
        start = time.time()
        limit = 100
        if "LIMIT" in words:
            words, limit = words.split("LIMIT")
            limit = int(limit)
        tokens = [keys.token_key(w) for w in preprocess_query(words)]
        DB.zinterstore(words, tokens)
        results = DB.zrevrange(words, 0, limit, withscores=True)
        DB.delete(words)
        for id_, score in results:
            r = Result(id_)
            print("{} {} {}".format(white(r), blue(r._id), cyan(score)))
        duration = round((time.time() - start) * 1000, 1)
        print(magenta("({} in {} ms)".format(len(results), duration)))

    def do_DBINFO(self, *args):
        """Print some useful infos from Redis DB."""
        info = DB.info()
        keys = [
            "keyspace_misses",
            "keyspace_hits",
            "used_memory_human",
            "total_commands_processed",
            "total_connections_received",
            "connected_clients",
        ]
        for key in keys:
            print("{}: {}".format(white(key), blue(info[key])))
        nb_of_redis_db = int(DB.config_get("databases")["databases"])
        for db_index in range(nb_of_redis_db - 1):
            db_name = "db{}".format(db_index)
            if db_name in info:
                label = white("nb keys (db {})".format(db_index))
                print("{}: {}".format(label, blue(info[db_name]["keys"])))

    def do_DBKEY(self, key):
        """Print raw content of a DB key.
        DBKEY g|u09tyzfe"""
        type_ = DB.type(key).decode()
        if type_ == "set":
            out = DB.smembers(key)
        elif type_ == "string":
            out = DB.get(key)
        else:
            out = "Unsupported type {}".format(type_)
        print("type:", magenta(type_))
        print("value:", white(out))

    def do_GEODISTANCE(self, s):
        """Compute geodistance from a result to a point.
        GEODISTANCE 772210180J 48.1234 2.9876"""
        try:
            _id, lat, lon = s.split()
        except:
            return self.error("Malformed query. Use: ID lat lon")
        try:
            result = Result(keys.document_key(_id))
        except ValueError as e:
            return self.error(e)
        center = (float(lat), float(lon))
        km = haversine_distance((float(result.lat), float(result.lon)), center)
        score = km_to_score(km)
        print("km: {} | score: {}".format(white(km), blue(score)))

    def do_GEOHASHTOGEOJSON(self, geoh):
        """Build GeoJSON corresponding to geohash given as parameter.
        GEOHASHTOGEOJSON u09vej04 [NEIGHBORS 0|1|2]"""
        geoh, with_neighbors = self._match_option("NEIGHBORS", geoh)
        bbox = geohash.bbox(geoh)
        try:
            with_neighbors = int(with_neighbors)
        except TypeError:
            with_neighbors = 0

        def expand(bbox, geoh, depth):
            neighbors = geohash.neighbors(geoh)
            for neighbor in neighbors:
                other = geohash.bbox(neighbor)
                if with_neighbors > depth:
                    expand(bbox, neighbor, depth + 1)
                else:
                    if other["n"] > bbox["n"]:
                        bbox["n"] = other["n"]
                    if other["s"] < bbox["s"]:
                        bbox["s"] = other["s"]
                    if other["e"] > bbox["e"]:
                        bbox["e"] = other["e"]
                    if other["w"] < bbox["w"]:
                        bbox["w"] = other["w"]

        if with_neighbors > 0:
            expand(bbox, geoh, 0)

        geojson = {
            "type": "Polygon",
            "coordinates": [
                [
                    [bbox["w"], bbox["n"]],
                    [bbox["e"], bbox["n"]],
                    [bbox["e"], bbox["s"]],
                    [bbox["w"], bbox["s"]],
                    [bbox["w"], bbox["n"]],
                ]
            ],
        }
        print(white(json.dumps(geojson)))

    def do_GEOHASH(self, latlon):
        """Compute a geohash from latitude and longitude.
        GEOHASH 48.1234 2.9876"""
        try:
            lat, lon = map(float, latlon.split())
        except ValueError:
            print(red("Invalid lat and lon {}".format(latlon)))
        else:
            print(white(geohash.encode(lat, lon, config.GEOHASH_PRECISION)))

    def do_GEOHASHMEMBERS(self, geoh):
        """Return members of a geohash and its neighbors.
        GEOHASHMEMBERS u09vej04 [NEIGHBORS 0]"""
        geoh, with_neighbors = self._match_option("NEIGHBORS", geoh)
        key = compute_geohash_key(geoh, with_neighbors != "0")
        if key:
            for id_ in DB.smembers(key):
                r = Result(id_)
                print("{} {}".format(white(r), blue(r._id)))

    def do_GET(self, _id):
        """Get document from index with its id.
        GET 772210180J"""
        doc = doc_by_id(_id)
        if not doc:
            return self.error('id "{}" not found'.format(_id))
        for key, value in doc.items():
            if key == config.HOUSENUMBERS_FIELD:
                continue
            print("{} {}".format(white(key), magenta(value)))
        if doc.get("housenumbers"):

            def sorter(v):
                try:
                    return int(re.match(r"^\d+", v["raw"]).group())
                except AttributeError:
                    return -1

            housenumbers = sorted(doc["housenumbers"].values(), key=sorter)
            print(
                white("housenumbers"),
                magenta(", ".join(v["raw"] for v in housenumbers)),
            )

    def do_FREQUENCY(self, word):
        """Return word frequency in index.
        FREQUENCY lilas"""
        print(white(word_frequency(word)))

    def _print_field_index_details(self, field, _id):
        for token in indexed_string(field):
            print(
                white(token),
                blue(DB.zscore(keys.token_key(token), keys.document_key(_id))),
                blue(DB.zrevrank(keys.token_key(token), keys.document_key(_id))),
            )

    def do_INDEX(self, _id):
        """Get index details for a document by its id.
        INDEX 772210180J"""
        doc = doc_by_id(_id)
        if not doc:
            return self.error('id "{}" not found'.format(_id))
        for field in config.FIELDS:
            key = field["key"]
            if key in doc:
                self._print_field_index_details(doc[key], _id)

    def do_BESTSCORE(self, word):
        """Return document linked to word with higher score.
        BESTSCORE lilas"""
        key = keys.token_key(indexed_string(word)[0])
        for _id, score in DB.zrevrange(key, 0, 20, withscores=True):
            result = Result(_id)
            print(white(result), blue(score), green(result._id))

    def do_REVERSE(self, latlon):
        """Do a reverse search. Args: lat lon.
        REVERSE 48.1234 2.9876"""
        lat, lon = latlon.split()
        for r in reverse(float(lat), float(lon)):
            print(
                "{} ({} | {} km | {})".format(
                    white(r), blue(r.score), blue(r.distance), blue(r._id)
                )
            )

    def do_TOKENIZE(self, string):
        """Inspect how a string is tokenized before being indexed.
        TOKENIZE Rue des Lilas"""
        print(white(" ".join(indexed_string(string))))

    def do_STRDISTANCE(self, s):
        """Print the distance score between two strings. Use | as separator.
        STRDISTANCE rue des lilas|porte des lilas"""
        s = s.split("|")
        if not len(s) == 2:
            print(red("Malformed string. Use | between the two strings."))
            return
        one, two = s
        print(white(compare_str(one, two)))

    def do_CONFIG(self, name):
        """Inspect loaded Addok config. Output all config without argument.
        CONFIG [CONFIG_KEY]"""
        if not name:
            for name in self.complete_CONFIG():
                self.do_CONFIG(name)
            return
        value = getattr(config, name.upper(), "Not found.")
        print(blue(name), white(format_config(value)))

    def complete_CONFIG(self, text=None, *ignored):
        text = text or ""
        return [a for a in config.keys() if a.startswith(text) and a.isupper()]

    def do_SCRIPT(self, args):
        """Run a Lua script. Takes the raw Redis arguments.
        SCRIPT script_name number_of_keys key1 key2… arg1 arg2
        """
        try:
            name, keys_count, *args = args.split()
        except ValueError:
            print(red("Not enough arguments"))
            return
        try:
            keys_count = int(keys_count)
        except ValueError:
            print(red("You must pass the number of keys as first argument"))
            self.do_HELP("SCRIPT")
            return
        keys = args[:keys_count]
        args = args[keys_count:]
        try:
            output = getattr(scripts, name)(keys=keys, args=args)
        except AttributeError:
            print(red("No script named {}".format(name)))
            return
        except DB.Error as e:
            print(red(e))
            return
        if not isinstance(output, list):
            # Script may return just an integer.
            output = [output]
        for line in output:
            print(white(line))


def format_config(value):
    out = ""
    if isinstance(value, (list, tuple)):
        for item in value:
            out += format_config(item)
    elif type(value).__name__ == "function":
        out = "\n{}.{}".format(str(value.__module__), value.__name__)
    else:
        out = "\n{}".format(value)
    return out


def invoke(args=None):
    cmd = Cmd()
    hooks.register_shell_command(cmd)
    cmd.loop()


def pyinvoke(args=None):
    try:
        from IPython import start_ipython
    except ImportError:
        print(red('Import is not installed. Type "pip install ipython"'))
    else:
        start_ipython(
            argv=[], user_ns={"DB": DB, "config": config, "get_document": get_document}
        )


def register_command(subparsers):
    parser = subparsers.add_parser("shell", help="Run a shell to inspect Addok")
    parser.set_defaults(func=invoke)
    parser = subparsers.add_parser(
        "pyshell", help="Run a python shell with Addok config"
    )
    parser.set_defaults(func=pyinvoke)


def doc_by_id(_id):
    return get_document(keys.document_key(_id).encode())


def indexed_string(s):
    return list(preprocess_query(s))


def word_frequency(word):
    try:
        token = list(preprocess_query(word))[0]
    except IndexError:
        # Word has been filtered out.
        return
    return token_frequency(token)


def set_debug():
    logging.basicConfig(level=logging.DEBUG)



from collections import OrderedDict
from importlib.metadata import entry_points, PackageNotFoundError

plugins = OrderedDict()
blocked_plugins = set([])

def load():
    try:
        eps = entry_points()
        # Python ≥ 3.10: entry_points() returns an object with select() method
        # Python 3.9: entry_points() returns a dict-like object
        # In Python 3.14, the old dict-like API has been removed
        if hasattr(eps, "select"):
            selected = eps.select(group="addok.ext")
        else:  # Python 3.9
            selected = eps.get("addok.ext", [])
        for ep in selected:
            register(ep.load(), ep.name)
    except PackageNotFoundError:
        pass


def register(module, name=None):
    if name is None:
        name = module.__name__
    if name in blocked_plugins:
        print("Requested registration of plugin {name} but this plugin is blocked")
        return
    plugins[name] = module


def spec(func):
    def caller(*args, **kwargs):
        for plugin in plugins.copy().values():
            try:
                getattr(plugin, func.__name__)(*args, **kwargs)
            except AttributeError:
                pass

    return caller


def block(name_or_module, reason=""):
    if not isinstance(name_or_module, str):
        name_or_module = name_or_module.__name__
    print(f"Blocking plugin {name_or_module}: {reason}")
    if name_or_module in plugins:
        del plugins[name_or_module]
    blocked_plugins.add(name_or_module)


@spec
def register_http_endpoint(api):
    """Add new endpoints to Addok API."""


@spec
def register_http_middleware(middlewares):
    """Add new middlewares to Addok API."""


@spec
def preconfigure(config):
    """Configure addok by patching config object before user local config."""


@spec
def configure(config):
    """Configure addok by patching config object after user local config."""


@spec
def register_command(subparsers):
    """Register command for Addok CLI."""


@spec
def register_shell_command(cmd):
    """Register command for Addok shell."""



import pytest

from addok.helpers import import_by_path, load_csv_file
from addok.helpers.text import tokenize


@pytest.mark.parametrize(
    "input,expected",
    [
        ("addok.helpers.text.tokenize", tokenize),
        (tokenize, tokenize),
    ],
)
def test_import_by_path(input, expected):
    assert import_by_path(input) == expected


def test_import_csv_file():
    assert list(load_csv_file("tests/test.csv")) == [
        {"name": "foo", "city": "bar", "somethingelse": "baz"},
        {
            "name": "another",
            "city": "line",
            "somethingelse": "complete",
        },
    ]



import csv
import os
import pickle
import sys
from functools import wraps
from importlib import import_module
from math import asin, cos, exp, radians, sin, sqrt
from multiprocessing import get_context
from multiprocessing.pool import RUN, IMapUnorderedIterator, Pool
from types import FunctionType, ModuleType

from progressist import ProgressBar

from addok.config import config

PYTHON_VERSION = sys.version_info


def load_file(filepath):
    with open(filepath) as f:
        for line in f:
            yield line


def load_msgpack_file(filepath):
    import msgpack  # We don't want to make it a required dependency.

    with open(filepath, mode="rb") as f:
        for line in msgpack.Unpacker(f, encoding="utf-8"):
            yield line


def load_csv_file(filepath):
    with open(filepath, newline="") as csvfile:
        dialect = csv.Sniffer().sniff(csvfile.read(1024))
        csvfile.seek(0)
        reader = csv.DictReader(csvfile, dialect=dialect)
        yield from reader


def iter_pipe(pipe, processors):
    """Allow for iterators to return either an item or an iterator of items."""
    if isinstance(pipe, str):
        pipe = [pipe]
    for it in processors:
        pipe = it(pipe)
    yield from pipe


def import_by_path(path):
    """
    Import functions or class by their path. Should be of the form:
    path.to.module.func
    """
    if not isinstance(path, str):
        return path
    module_path, *name = path.rsplit(".", 1)
    func = import_module(module_path)
    if name:
        func = getattr(func, name[0])
    return func


def yielder(func):
    @wraps(func)
    def wrapper(pipe):
        for item in pipe:
            yield func(item)

    return wrapper


def haversine_distance(point1, point2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees).
    """
    lat1, lon1 = point1
    lat2, lon2 = point2

    # Convert decimal degrees to radians.
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # Haversine formula.
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
    c = 2 * asin(sqrt(a))

    # 6367 km is the radius of the Earth.
    km = 6367 * c
    return km


def km_to_score(km):
    # Score between 0 and 0.1 (close to 0 km will be close to 0.1, and 100 and
    # above will be 0).
    return 0.0 if km > 100 else exp(-((km / 50.0) ** 2))


COLORS = {
    "red": "31",
    "green": "32",
    "yellow": "33",
    "blue": "34",
    "magenta": "35",
    "cyan": "36",
    "white": "37",
    "reset": "39",
}


def colorText(s, color):
    # color should be a string from COLORS
    return "\033[%sm%s\033[%sm" % (COLORS[color], s, COLORS["reset"])


def red(s):
    return colorText(s, "red")


def green(s):
    return colorText(s, "green")


def yellow(s):
    return colorText(s, "yellow")


def blue(s):
    return colorText(s, "blue")


def magenta(s):
    return colorText(s, "magenta")


def cyan(s):
    return colorText(s, "cyan")


def white(s):
    return colorText(s, "white")


class Bar(ProgressBar):
    animation = "{spinner}"
    template = "{prefix} {animation} Done: {done} | Elapsed: {elapsed}"


class ChunkedPool(Pool):
    def imap_unordered(self, func, iterable, chunksize):
        """Customized version of imap_unordered.

        Directly send chunks to func, instead of iterating in each process and
        sending one by one.

        Original:
        https://hg.python.org/cpython/file/tip/Lib/multiprocessing/pool.py#l271

        Other tried options:
        - map_async: makes a list(iterable), so it loads all the data for each
          process into RAM
        - apply_async: needs manual chunking
        """
        assert self._state == RUN
        task_batches = Pool._get_tasks(func, iterable, chunksize)
        result = IMapUnorderedIterator(
            self if PYTHON_VERSION >= (3, 8) else self._cache
        )
        tasks = (
            (result._job, i, func, chunk, {})
            for i, (_, chunk) in enumerate(task_batches)
        )
        self._taskqueue.put((tasks, result._set_length))
        return result


def _get_config_overrides():
    """Extract configuration values that differ from defaults.

    Returns:
        Dict of config keys and their overridden values (only simple serializable types).
    """
    # Import here to avoid circular dependencies
    from addok.config import default as default_config

    # Only propagate uppercase config keys (actual settings, not methods)
    overrides = {}
    for key in config.keys():
        if not key.isupper():
            continue

        current_value = config.get(key)
        default_value = getattr(default_config, key, None)

        # Skip if values are the same
        if current_value == default_value:
            continue

        # Skip non-serializable types (functions, classes, modules)
        if isinstance(current_value, (FunctionType, type, ModuleType)):
            continue

        # Skip lists/dicts containing functions or classes
        if isinstance(current_value, (list, tuple)):
            if any(isinstance(item, (FunctionType, type, ModuleType)) for item in current_value):
                continue
        elif isinstance(current_value, dict):
            if any(isinstance(v, (FunctionType, type, ModuleType)) for v in current_value.values()):
                continue

        # Test if value is pickle-serializable
        try:
            pickle.dumps(current_value)
            overrides[key] = current_value
        except (pickle.PicklingError, TypeError, AttributeError):
            # Skip values that can't be pickled
            pass

    return overrides


def _worker_init(redis_params, config_env_vars=None, config_overrides=None):
    """Initialize Redis connection in worker processes for multiprocessing.

    This is required for spawn context where the connection object doesn't
    get properly inherited by child processes.

    Args:
        redis_params: Dict with Redis connection parameters (host, port, db, etc.)
        config_env_vars: Dict of environment variables to set for configuration
        config_overrides: Dict of config values to override after loading
    """
    # Import here to avoid circular dependencies and ensure fresh imports in spawned processes
    from addok import ds
    from addok.config import config as addok_config
    from addok.db import DB

    # Set environment variables for configuration
    if config_env_vars:
        for key, value in config_env_vars.items():
            if value is not None:
                os.environ[key] = str(value)

    # Force reload of config in worker process
    if not addok_config.loaded:
        addok_config.load()

    # Apply config overrides (for test monkeypatching)
    if config_overrides:
        for key, value in config_overrides.items():
            addok_config[key] = value

    # Connect to Redis with provided parameters
    DB.connect(**redis_params['indexes'])

    # Also connect document store if using Redis
    if 'documents' in redis_params and redis_params.get('use_redis_documents'):
        ds._DB.connect(**redis_params['documents'])


def parallelize(func, iterable, chunk_size, **bar_kwargs):
    """Execute func on iterable chunks using multiprocessing.

    Uses 'spawn' context on macOS for fork-safety, 'fork' on Linux for speed.
    """
    # Import here to avoid circular dependencies
    from addok.db import get_redis_params

    bar = Bar(prefix="Processing…", **bar_kwargs)

    # Prepare worker initialization parameters
    redis_params = get_redis_params()
    config_env_vars = {'ADDOK_CONFIG_MODULE': os.environ.get('ADDOK_CONFIG_MODULE')} if 'ADDOK_CONFIG_MODULE' in os.environ else {}
    config_overrides = _get_config_overrides()

    # Use 'spawn' on macOS for fork-safety, 'fork' on Linux for performance
    context = get_context('spawn' if sys.platform == 'darwin' else 'fork')

    with ChunkedPool(
        processes=config.BATCH_WORKERS,
        initializer=_worker_init,
        initargs=(redis_params, config_env_vars, config_overrides),
        context=context
    ) as pool:
        for chunk in pool.imap_unordered(func, iterable, chunk_size):
            bar(step=len(chunk))
        bar.finish()



import hashlib
import uuid
import time

import geohash

from .config import config
from .db import DB
from .ds import get_document, get_documents
from .helpers import keys as dbkeys, scripts
from .helpers.text import ascii

REDIS_UNIQUE_ID = str(uuid.uuid4())  # Really unique id for tmp values in redis.


def compute_geohash_key(geoh, with_neighbors=True):
    if with_neighbors:
        neighbors = geohash.expand(geoh)
        neighbors = [dbkeys.geohash_key(n) for n in neighbors]
    else:
        neighbors = [geoh]
    key = "gx|{}".format(geoh)
    total = DB.sunionstore(key, neighbors)
    if not total:
        # No need to keep it.
        DB.delete(key)
        key = False
    else:
        DB.expire(key, 10)
    return key


class Result:
    def __init__(self, _id):
        self.housenumber = None
        self._scores = {}
        self.load(_id)
        self.labels = []

    def load(self, doc_or_id):
        self._cache = {}
        if isinstance(doc_or_id, dict):
            doc = doc_or_id
        else:
            doc = get_document(doc_or_id)
        if not doc:
            raise ValueError('id "{}" not found'.format(doc_or_id[2:]))
        self._doc = doc

    def __getattr__(self, key):
        if key == "_id":
            # result._id should load the id whatever the real field used.
            key = config.ID_FIELD
        if key not in self._cache:
            # By convention, in case of multiple values, first value is default
            # value, others are aliases.
            value = self._rawattr(key)[0]
            self._cache[key] = value
        return self._cache[key]

    def __str__(self):
        return (
            str(self.labels[0]) if self.labels else self._rawattr(config.NAME_FIELD)[0]
        )

    def _rawattr(self, key):
        value = self._doc.get(key, "")
        if not isinstance(value, (tuple, list)):
            value = [value]
        return value

    def __repr__(self):
        return "<{} - {} ({})>".format(str(self), self._id, self.score)

    @property
    def keys(self):
        keys = ["housenumber"] + list(self._doc.keys())
        # housenumbers are too verbose for a given street.
        yield from (key for key in keys if key != "housenumbers")

    def update(self, data):
        self._doc.update(data)
        self._cache.update(data)

    def format(self):
        result = self
        for formatter in config.RESULTS_FORMATTERS:
            result = formatter(result)
        return result

    def add_score(self, name, score, ceiling):
        if score >= self._scores.get(name, (0, 0))[0]:
            self._scores[name] = (score, ceiling)

    @property
    def score(self):
        if self._score != "":
            return float(self._score)
        score, _max = zip(*self._scores.values())
        return sum(score) / sum(_max)

    @score.setter
    def score(self, value):
        self._score = value

    @property
    def str_distance(self):
        return self._scores.get("str_distance", [0.0])[0]

    @classmethod
    def from_id(self, _id):
        """Return a result from it's document _id."""
        return Result(dbkeys.document_key(_id))


class BaseHelper:
    def __init__(self, verbose):
        self._start = time.time()
        self._debug = []
        if not verbose:
            self.debug = lambda *args: None

    def debug(self, *args):
        s = args[0] % args[1:]
        s = "[{}] {}".format(str((time.time() - self._start) * 1000)[:5], s)
        self._debug.append(s)

    def report(self):  # pragma: no cover
        for msg in self._debug:
            print(msg)

    def _normalize_filter_value(self, value):
        """Normalize filter value by removing spaces, sorting, and deduplicating.

        Args:
            value: Raw filter value string (e.g., "street + city  +street")

        Returns:
            Normalized string with values sorted and deduplicated (e.g., "city+street")
        """
        cleaned = value.replace(' ', '+').strip()
        # Split, remove empty strings, deduplicate, limit to config.MAX_FILTER_VALUES
        values = [v for v in cleaned.split('+') if v]
        unique_values = list(dict.fromkeys(values))[:config.MAX_FILTER_VALUES]
        return '+'.join(sorted(unique_values))

    def _compute_multifilter(self, filter_name, normalized_value):
        """Create temporary OR filter keys using Redis SUNIONSTORE.

        This combines multiple filter values with OR logic. For example,
        type=street+city will match documents where type is "street" OR "city".

        The resulting key is cached in Redis with an expiration time, or persisted
        if the result set is large (>100k items).

        Args:
            filter_name: The filter field name (e.g., "type")
            normalized_value: Normalized filter value (e.g., "city+street")

        Returns:
            The Redis key for the combined filter
        """
        key = dbkeys.filter_key(filter_name, normalized_value)

        if not DB.exists(key):
            self.debug(f'MultiFilter created: {filter_name}={normalized_value}')
            keys = [dbkeys.filter_key(filter_name, v)
                    for v in normalized_value.split('+')]
            DB.sunionstore(key, keys)

        # Persist large filters to avoid recomputation, expire small ones
        if DB.scard(key) > 100000:
            DB.persist(key)
            self.debug(f'MultiFilter persistent: {filter_name}={normalized_value}')
        else:
            DB.expire(key, 10)

        return key

    def _build_filters(self, filters):
        """Build filter keys with normalized multi-value support.

        This method processes filter parameters, normalizes multi-value filters,
        creates temporary Redis keys for OR operations, and optionally combines
        multiple filters with AND logic.

        Args:
            filters: Dictionary of filter parameters (e.g., {"type": "street+city"})

        Returns:
            List of filter keys to use in queries
        """
        # Build filter keys with normalized multi-value support
        filter_keys = []
        for k, v in filters.items():
            if v.strip():
                normalized = self._normalize_filter_value(v)
                # Check if this is a multi-value filter
                if '+' in normalized:
                    # Create temporary OR filter key
                    filter_key = self._compute_multifilter(k, normalized)
                else:
                    # Single value filter
                    filter_key = dbkeys.filter_key(k, normalized)
                filter_keys.append(filter_key)

        # Combine multiple filters with AND logic
        if len(filter_keys) > 1:
            return self._combine_filters_with_keys(filter_keys)

        return filter_keys

    def _combine_filters_with_keys(self, filter_keys):
        """Combine filter keys with AND logic using Redis SINTERSTORE.

        Args:
            filter_keys: List of filter keys to combine

        Returns:
            List containing a single combined filter key
        """
        # Use a stable hash instead of repr() for the cache key
        filter_string = '|'.join(sorted(filter_keys))
        key_hash = hashlib.md5(filter_string.encode()).hexdigest()
        key = f"combined:{key_hash}"

        if not DB.exists(key):
            self.debug(f'Combined filter: {filter_string}')
            DB.sinterstore(key, filter_keys)

        DB.expire(key, 10)
        return [key]


class Search(BaseHelper):

    MAX_MEANINGFUL = 10

    def __init__(self, fuzzy=1, limit=10, autocomplete=True, verbose=False):
        super().__init__(verbose=verbose)
        self.fuzzy = fuzzy
        self.wanted = limit
        self.autocomplete = autocomplete
        self.pid = REDIS_UNIQUE_ID

    def __call__(self, query, lat=None, lon=None, **filters):
        self.lat = lat
        self.lon = lon
        self._geohash_key = None
        self.results = {}
        self.bucket = set([])  # No duplicates.
        self.meaningful = []
        self.not_found = []
        self.common = []
        self.housenumbers = []
        self.keys = []
        self.matched_keys = set([])
        self.check_housenumber = filters.get("type") in [None, "housenumber"]
        self.only_housenumber = filters.get("type") == "housenumber"
        # Build filter keys with normalized multi-value support
        self.filters = self._build_filters(filters)

        self.query = ascii(query.strip())
        for func in config.SEARCH_PREPROCESSORS:
            func(self)
        self.debug("Taken tokens: %s", self.meaningful)
        self.debug("Common tokens: %s", self.common)
        self.debug("Housenumbers token: %s", self.housenumbers)
        self.debug("Not found tokens: %s", self.not_found)

        self.debug('Filters: %s', [f'{k}={v}' for k, v in filters.items()])

        for collector in config.RESULTS_COLLECTORS:
            self.debug("** %s **", collector.__name__.upper())
            if collector(self):
                break
        return list(self.render())

    @property
    def geohash_key(self):
        if self.lat and self.lon and self._geohash_key is None:
            geoh = geohash.encode(self.lat, self.lon, config.GEOHASH_PRECISION)
            self._geohash_key = compute_geohash_key(geoh)
            if self._geohash_key:
                self.debug("Computed geohash key %s", self._geohash_key)
            else:
                self.debug("Empty geohash key, deleting %s", self._geohash_key)
        return self._geohash_key

    def render(self):
        self.convert()
        self._sorted_bucket = list(self.results.values())
        self._sorted_bucket.sort(key=lambda r: r.score, reverse=True)
        for result in self._sorted_bucket[: self.wanted]:
            if result.score < config.MIN_SCORE:
                self.debug("Score too low (%s), removing `%s`", result.score, result)
                continue
            yield result

    def intersect(self, keys, limit=0):
        if not limit > 0:
            limit = max(self.wanted, config.BUCKET_MAX)
        ids = []
        if keys:
            if self.filters:
                keys.extend(self.filters)
            if len(keys) == 1:
                key = keys[0]
                if key.startswith(dbkeys.TOKEN_PREFIX):
                    ids = DB.zrevrange(key, 0, limit - 1)
                else:
                    ids = DB.smembers(key)
            else:
                ids = scripts.zinter(keys=set(keys), args=[self.pid, limit])
        return set(ids)

    def add_to_bucket(self, keys, limit=None):
        self.debug("Adding to bucket with keys %s", keys)
        self.matched_keys.update([k for k in keys if k.startswith(dbkeys.TOKEN_PREFIX)])
        limit = limit or (config.BUCKET_MAX - len(self.bucket))
        self.bucket.update(self.intersect(keys, limit))
        self.debug("%s ids in bucket so far", len(self.bucket))

    def new_bucket(self, keys, limit=0):
        self.debug("New bucket with keys %s and limit %s", keys, limit)
        self.matched_keys = set([k for k in keys if k.startswith(dbkeys.TOKEN_PREFIX)])
        self.bucket = self.intersect(keys, limit)
        self.debug("%s ids in bucket so far", len(self.bucket))

    def convert(self):
        self.debug("Computing results")
        ids = [i for i in self.bucket if i not in self.results]
        if ids:
            documents = get_documents(*ids)
            self.debug("Done getting results data")
            for _id, doc in documents:
                result = Result(doc)
                for processor in config.SEARCH_RESULT_PROCESSORS:
                    valid = processor(self, result)
                    if valid is False:
                        break
                else:
                    self.results[_id] = result
        self.debug("Done computing results")

    @property
    def bucket_full(self):
        l = len(self.bucket)
        return l >= self.wanted and l < config.BUCKET_MAX

    @property
    def bucket_overflow(self):
        return len(self.bucket) >= config.BUCKET_MAX

    @property
    def bucket_dry(self):
        return len(self.bucket) < self.wanted

    @property
    def bucket_empty(self):
        return not self.bucket

    @property
    def cream(self):
        return len(
            [
                r
                for _id, r in self.results.items()
                if r.str_distance >= config.MATCH_THRESHOLD
            ]
        )

    def has_cream(self):
        if (
            self.bucket_empty
            or self.bucket_overflow
            or len(self.bucket) > config.BUCKET_MIN
        ):
            return False
        self.debug("Checking cream.")
        self.convert()
        return self.cream > 0

    @property
    def pass_should_match_threshold(self):
        return len(self.matched_keys) >= self.should_match_threshold

    @property
    def only_commons(self):
        return self.tokens and len(self.tokens) == len(self.common)


class Reverse(BaseHelper):
    def __call__(self, lat, lon, limit=1, **filters):
        self.lat = lat
        self.lon = lon
        self.keys = set([])
        self.results = []
        self.wanted = limit
        self.fetched = []
        self.check_housenumber = filters.get("type") in [None, "housenumber"]
        self.only_housenumber = filters.get("type") == "housenumber"
        # Build filter keys with normalized multi-value support
        self.filters = self._build_filters(filters)
        self.debug('Filters: %s', [f'{k}={v}' for k, v in filters.items()])
        geoh = geohash.encode(lat, lon, config.GEOHASH_PRECISION)
        hashes = self.expand([geoh])
        self.fetch(hashes)
        if not self.keys:
            hashes = self.expand(hashes)
            self.fetch(hashes)
        return self.convert()

    def expand(self, hashes):
        new = []
        for h in hashes:
            neighbors = geohash.expand(h)
            for n in neighbors:
                if n not in self.fetched:
                    new.append(n)
        return new

    def fetch(self, hashes):
        self.debug("Fetching %s", hashes)
        for h in hashes:
            k = dbkeys.geohash_key(h)
            self.intersect(k)
            self.fetched.append(h)

    def intersect(self, key):
        if self.filters:
            keys = DB.sinter([key] + self.filters)
        else:
            keys = DB.smembers(key)
        self.keys.update(keys)

    def convert(self):
        for _id in self.keys:
            result = Result(_id)
            for processor in config.REVERSE_RESULT_PROCESSORS:
                valid = processor(self, result)
                if valid is False:
                    self.debug("Result removed by processor: %s", result)
                    break
            else:
                self.results.append(result)
                self.debug(result, result.distance, result.score)
        self.results.sort(key=lambda r: r.score, reverse=True)
        return self.results[: self.wanted]


def search(
    query,
    fuzzy=1,
    limit=10,
    autocomplete=False,
    lat=None,
    lon=None,
    verbose=False,
    **filters
):
    helper = Search(
        fuzzy=fuzzy,
        limit=limit,
        verbose=verbose,
        autocomplete=autocomplete,
    )
    return helper(query, lat=lat, lon=lon, **filters)


def reverse(lat, lon, limit=1, verbose=False, **filters):
    helper = Reverse(verbose=verbose)
    return helper(lat, lon, limit, **filters)



import json
import os.path
import sys
from datetime import timedelta

from addok.config import config
from addok.db import DB
from addok.ds import DS
from addok.helpers import iter_pipe, parallelize, yielder


def run(args):
    if args.filepath:
        for path in args.filepath:
            process_file(path)
    elif not sys.stdin.isatty():  # Any better way to check for stdin?
        process_stdin(sys.stdin)


def reset(args):
    if args.force or input('Type "yes" to delete ALL data: ') == "yes":
        DB.flushdb()
        DS.flushdb()
        print("All data has been deleted.")
    else:
        print("Nothing has been deleted.")


def register_command(subparsers):
    parser = subparsers.add_parser("batch", help="Batch import documents")
    parser.add_argument("filepath", nargs="*", help="Path to file to process")
    parser.set_defaults(func=run)
    parser = subparsers.add_parser("reset", help="Delete ALL indexes and documents")
    parser.add_argument("--force", help="Do not ask for confirm", action="store_true")
    parser.set_defaults(func=reset)


def process_file(filepath):
    print("Import from file", filepath)
    _, ext = os.path.splitext(filepath)
    if not os.path.exists(filepath):
        sys.stderr.write("File not found: {}".format(filepath))
        sys.exit(1)
    config.INDEX_EDGE_NGRAMS = False  # Run command "ngrams" instead.
    batch(config.BATCH_FILE_LOADER(filepath))


def process_stdin(stdin):
    print("Import from stdin")
    batch(stdin)


@yielder
def to_json(row):
    try:
        return json.loads(row)
    except ValueError:
        return None


def process_documents(*docs):
    return list(iter_pipe(docs, config.BATCH_PROCESSORS))


def batch(iterable):
    parallelize(
        process_documents,
        iterable,
        chunk_size=config.BATCH_CHUNK_SIZE,
        throttle=timedelta(seconds=1),
    )



import redis

from addok.config import config
from addok.db import DB
from addok.helpers import keys as dbkeys
from addok.helpers import magenta, parallelize, white, scripts
from addok.helpers.index import token_key_frequency
from addok.helpers.search import preprocess_query
from addok.helpers.text import compute_edge_ngrams
from addok.pairs import pair_key


def edge_ngram_key(s):
    return "n|{}".format(s)


def index_edge_ngrams(pipe, token):
    for ngram in compute_edge_ngrams(token):
        pipe.sadd(edge_ngram_key(ngram), token)


def deindex_edge_ngrams(token):
    for ngram in compute_edge_ngrams(token):
        DB.srem(edge_ngram_key(ngram), token)


class EdgeNgramIndexer:
    @staticmethod
    def index(pipe, key, doc, tokens, **kwargs):
        if config.INDEX_EDGE_NGRAMS:  # Allow to disable for mass indexing.
            for token in tokens.keys():
                index_edge_ngrams(pipe, token)

    @staticmethod
    def deindex(db, key, doc, tokens, **kwargs):
        if config.INDEX_EDGE_NGRAMS:
            for token in tokens:
                tkey = dbkeys.token_key(token)
                if not DB.exists(tkey):
                    deindex_edge_ngrams(token)


def only_commons_but_geohash_try_autocomplete_collector(helper):
    if helper.geohash_key and helper.only_commons:
        autocomplete(helper, helper.tokens, use_geohash=True)


def only_commons_try_autocomplete_collector(helper):
    if helper.only_commons:
        autocomplete(helper, helper.tokens, skip_commons=True)
        if not helper.bucket_empty:
            helper.debug("Only common terms. Return.")
            return True


def no_meaningful_but_common_try_autocomplete_collector(helper):
    if not helper.meaningful and helper.common:
        # Only commons terms, try to reduce with autocomplete.
        helper.debug("Only commons, trying autocomplete")
        autocomplete(helper, helper.common)
        helper.meaningful = helper.common[:1]
        if not helper.pass_should_match_threshold:
            return False
        if helper.bucket_full or helper.bucket_overflow or helper.has_cream():
            return True


def autocomplete_meaningful_collector(helper):
    if helper.bucket_overflow:
        return
    if not helper.autocomplete:
        helper.debug("Autocomplete not active. Abort.")
        return
    if helper.geohash_key:
        autocomplete(helper, helper.meaningful, use_geohash=True)
    autocomplete(helper, helper.meaningful)


def autocomplete(helper, tokens, skip_commons=False, use_geohash=False):
    helper.debug("Autocompleting %s", helper.last_token)
    keys = [t.db_key for t in tokens if not t.is_last]
    pair_keys = [pair_key(t) for t in tokens if not t.is_last]
    key = edge_ngram_key(helper.last_token)
    autocomplete_tokens = DB.sinter(pair_keys + [key])
    if not autocomplete_tokens:
        helper.debug("No candidates. Aborting.")
        return
    token_keys = [dbkeys.token_key(t.decode()) for t in autocomplete_tokens]
    if len(tokens) == 1:
        helper.debug("Ordering candidates by max score")
        autocomplete_tokens = scripts.order_by_max_score(keys=token_keys)
    else:
        helper.debug("Ordering candidates by frequency")
        autocomplete_tokens = scripts.order_by_frequency(keys=token_keys)
    helper.debug(
        "Found tokens to autocomplete [%s, …]", b", ".join(autocomplete_tokens[:10])
    )
    for token in autocomplete_tokens:
        key = token.decode()
        if skip_commons and token_key_frequency(key) > config.COMMON_THRESHOLD:
            helper.debug("Skip common token to autocomplete %s", key)
            continue
        if not helper.bucket_overflow or helper.last_token in helper.not_found:
            helper.debug("Trying to extend bucket. Autocomplete %s", key)
            extra_keys = [key]
            if use_geohash and helper.geohash_key:
                extra_keys.append(helper.geohash_key)
            helper.add_to_bucket(keys + extra_keys)


def index_ngram_keys(*keys):
    pipe = DB.pipeline(transaction=False)
    for key in keys:
        key = key.decode()
        _, token = key.split("|")
        if token.isdigit():
            continue
        index_edge_ngrams(pipe, token)
    try:
        pipe.execute()
    except redis.RedisError as e:
        msg = "Error while generating ngrams:\n{}".format(str(e))
        raise ValueError(msg)
    return keys


def create_edge_ngrams(*args):
    pattern = "{}*".format(dbkeys.TOKEN_PREFIX)
    parallelize(
        index_ngram_keys, DB.scan_iter(match=pattern), chunk_size=10000, throttle=1000
    )


def register_command(subparsers):
    parser = subparsers.add_parser("ngrams", help="Create edge ngrams.")
    parser.set_defaults(func=create_edge_ngrams)


def do_AUTOCOMPLETE(cmd, s):
    """Shows autocomplete results for a given token."""
    s = list(preprocess_query(s))[0]
    keys = [k.decode() for k in DB.smembers(edge_ngram_key(s))]
    print(white(keys))
    print(magenta("({} elements)".format(len(keys))))


def register_shell_command(cmd):
    cmd.register_command(do_AUTOCOMPLETE)



import json
import zlib


class ZlibSerializer:
    @classmethod
    def dumps(cls, data):
        return zlib.compress(json.dumps(data).encode())

    @classmethod
    def loads(cls, data):
        return json.loads(zlib.decompress(data).decode())



from addok.config import config
from addok.db import DB, RedisProxy
from addok.helpers import keys


class RedisStore:
    def fetch(self, *keys):
        pipe = _DB.pipeline(transaction=False)
        for key in keys:
            pipe.get(key)
        for key, doc in zip(keys, pipe.execute()):
            if doc is not None:
                yield key, doc

    def upsert(self, *docs):
        pipe = _DB.pipeline(transaction=False)
        for key, blob in docs:
            pipe.set(key, blob)
        pipe.execute()

    def remove(self, *keys):
        pipe = _DB.pipeline(transaction=False)
        for key in keys:
            pipe.delete(key)
        pipe.execute()

    def flushdb(self):
        _DB.flushdb()


class DSProxy:
    instance = None

    def __getattr__(self, name):
        return getattr(self.instance, name)


_DB = RedisProxy()
DS = DSProxy()


@config.on_load
def on_load():
    DS.instance = config.DOCUMENT_STORE()
    # Do not create connection if not using this store class.
    if config.DOCUMENT_STORE == RedisStore:
        params = config.REDIS.copy()
        params.update(config.REDIS.get("documents", {}))
        _DB.connect(
            host=params.get("host"),
            port=params.get("port"),
            db=params.get("db"),
            password=params.get("password"),
            unix_socket_path=params.get("unix_socket_path"),
        )


def store_documents(docs):
    to_upsert = []
    to_remove = []
    for doc in docs:
        if not doc:
            continue
        if config.ID_FIELD not in doc:
            doc[config.ID_FIELD] = DB.next_id()
        key = keys.document_key(doc[config.ID_FIELD])
        if doc.get("_action") in ["delete", "update"]:
            to_remove.append(key)
        if doc.get("_action") in ["index", "update", None]:
            to_upsert.append((key, config.DOCUMENT_SERIALIZER.dumps(doc)))
        yield doc
    if to_remove:
        DS.remove(*to_remove)
    if to_upsert:
        DS.upsert(*to_upsert)


def get_document(key):
    results = DS.fetch(key)
    try:
        _, doc = next(results)
    except StopIteration:
        return None
    return config.DOCUMENT_SERIALIZER.loads(doc)


def get_documents(*keys):
    for id_, blob in DS.fetch(*keys):
        yield id_, config.DOCUMENT_SERIALIZER.loads(blob)



import re
from functools import lru_cache
from pathlib import Path

import editdistance
from ngram import NGram
from unidecode import unidecode

from addok.config import config
from addok.db import DB
from addok.helpers import keys, yielder
from addok.helpers.index import token_frequency


PATTERN = re.compile(r"[\w]+", re.U | re.X)


class Token(str):

    __slots__ = (
        "_positions",
        "is_last",
        "db_key",
        "raw",
        "_frequency",
        "_key",
        "kind",
        "is_first",
    )

    def __new__(cls, value, position=None, is_last=False, raw=None, kind=None):
        obj = str.__new__(cls, value)
        obj._positions = []
        position = position or 0
        obj.position = position
        obj.is_first = obj.position[0] == 0
        obj.is_last = is_last
        obj.db_key = None
        obj.raw = raw or value  # Allow to keep raw on update.
        obj.kind = kind
        return obj

    def __repr__(self):
        return "<Token {}>".format(self)

    def update(self, value, **kwargs):
        default = dict(
            is_last=self.is_last,
            raw=self.raw,
            kind=self.kind,
            position=self.position[:],
        )
        # Never replace position through `update`.
        position = kwargs.pop("position", None)
        default.update(kwargs)
        token = Token(value=value, **default)
        if position is not None:
            token.position = position
        return token

    def search(self):
        if DB.exists(self.key):
            self.db_key = self.key

    @property
    def is_common(self):
        return self.frequency > config.COMMON_THRESHOLD

    @property
    def frequency(self):
        if not hasattr(self, "_frequency"):
            self._frequency = token_frequency(self)
        return self._frequency

    @property
    def key(self):
        if not hasattr(self, "_key"):
            self._key = keys.token_key(self)
        return self._key

    @property
    def position(self):
        # Allow to store multiple positions when a token is subdivided.
        return self._positions

    @position.setter
    def position(self, position):
        if isinstance(position, list):
            self._positions = position
        else:
            self._positions.append(position)


def _tokenize(text):
    """Split text into a list of tokens."""
    return PATTERN.findall(text)


def tokenize(pipe):
    for text in pipe:
        for position, token in enumerate(_tokenize(text)):
            yield Token(token, position=position)


def _normalize(s):
    return s.update(unidecode(s.lower()))


normalize = yielder(_normalize)


@config.on_load
def load_synonyms():
    config.SYNONYMS = {}
    for path in config.SYNONYMS_PATHS:
        with Path(path).open() as f:
            for line in f:
                if line.startswith("#"):
                    continue
                synonyms, wanted = line.split("=>")
                wanted = wanted.strip()
                synonyms = synonyms.split(",")
                for synonym in synonyms:
                    synonym = synonym.strip()
                    if not synonym:
                        continue
                    config.SYNONYMS[synonym] = wanted


def synonymize(tokens):
    for token in tokens:
        for position, subtoken in enumerate(config.SYNONYMS.get(token, token).split()):
            yield token.update(subtoken, position=position)


class ascii(str):
    """Just like a str, but ascii folded and cached."""

    __slots__ = ["_cache", "_raw"]

    def __new__(cls, value):
        try:
            cache = value._cache
        except AttributeError:
            cache = alphanumerize(unidecode(value.lower()))
            obj = str.__new__(cls, cache)
            obj._cache = cache
            obj._raw = getattr(value, "_raw", value)
            return obj
        else:
            return value

    def __str__(self):
        return self._raw


@lru_cache(maxsize=512)
def ngrams(text, n=3):
    # Make sure strings are at least 3 chars long, and a given token will have
    # same ngrams whether at the start, the middle or the end of the string.
    text = " " + text + " "
    return set([text[i : i + n] for i in range(0, len(text) - (n - 1))])


def compare_ngrams(left, right, N=2, pad_len=0):
    left = ascii(left)
    right = ascii(right)
    if len(left) == 1 and len(right) == 1:
        # NGram.compare returns 0.0 for 1 letter comparison, even if letters
        # are equal.
        return 1.0 if left == right else 0.0
    return NGram.compare(left, right, N=N, pad_len=pad_len)


def compare_str(left, right):
    left = ascii(left)
    right = ascii(right)
    left_n = ngrams(left)
    right_n = ngrams(right)
    # Evaluate editdistance limited to common text portion.
    distance = (editdistance.eval(left, right) - abs(len(left) - len(right))) / max(
        len(left), len(right)
    )
    return (
        len(left_n & right_n) / len(right_n) * 0.85
        + len(left_n & right_n) / len(left_n) * 0.05
        + (1 - distance) * 0.1
    )


def contains(candidate, target):
    candidate = ascii(candidate)
    target = ascii(target)
    return candidate in target


def startswith(candidate, target):
    candidate = ascii(candidate)
    target = ascii(target)
    return target.startswith(candidate)


def equals(candidate, target):
    candidate = ascii(candidate)
    target = ascii(target)
    return target == candidate


def alphanumerize(text):
    return re.sub(r" {2,}", " ", re.sub(r"[^\w]", " ", text)).strip()


def compute_edge_ngrams(token, min=None):
    """Compute edge ngram of token from min. Does not include token itself."""
    if min is None:
        min = config.MIN_EDGE_NGRAMS
    token = token[: config.MAX_EDGE_NGRAMS + 1]
    return [token[:i] for i in range(min, len(token))]


class EntityTooLarge(ValueError):
    pass


@yielder
def check_query_length(q):
    if len(q) > config.QUERY_MAX_LENGTH:
        raise EntityTooLarge(
            "Query too long, {} chars, limit is {}".format(
                len(q), config.QUERY_MAX_LENGTH
            )
        )
    return q


@yielder
def flag_housenumber(token):
    """Very basic housenumber flagging. Make your own for your specific needs.
    Eg. in addok-france:
    https://github.com/addok/addok-france/blob/master/addok_france/utils.py#L106
    """
    if token.is_first and token.isdigit():
        token.kind = "housenumber"
    return token



from pathlib import Path
import os

import pytest

from addok import db, ds


def test_connection_port_is_not_default_one():
    assert db.DB.connection_pool.connection_kwargs["db"] == 14
    assert ds._DB.connection_pool.connection_kwargs["db"] == 15


def test_config_on_load_is_called_on_config_load():
    from addok.config import Config

    config = Config()

    @config.on_load
    def on_load():
        on_load.called = True

    config.load()
    assert on_load.called

    on_load.called = False

    # config should not reload if already loaded.
    config.load()
    assert not on_load.called


def test_local_config_has_been_loaded(config):
    # See addok/config/test.py
    assert config.COMMON_THRESHOLD == 1000


def test_config_does_not_fail_if_local_path_does_not_exist(capsys):
    os.environ["ADDOK_CONFIG_MODULE"] = "dummy/path.py"
    from addok.config import Config

    config = Config()
    config.load()
    out, err = capsys.readouterr()
    assert "No local config file found" in out
    os.environ["ADDOK_CONFIG_MODULE"] = ""


def test_config_load_exit_if_local_file_is_invalid():
    path = str(Path(__file__).parent / "invalid_config.py")
    os.environ["ADDOK_CONFIG_MODULE"] = path
    from addok.config import Config

    config = Config()
    with pytest.raises(SystemExit) as err:
        config.load()
        assert "Unable to import" in err
    os.environ["ADDOK_CONFIG_MODULE"] = ""


def test_config_on_load_consume_env_vars():
    from addok.config import Config

    os.environ["ADDOK_BATCH_WORKERS"] = "13"
    config = Config()

    config.load()
    assert config.BATCH_WORKERS == 13



"""Tests for Redis database helpers."""
from addok.db import _extract_redis_config, get_redis_params


def test_extract_redis_config_with_all_params():
    """Test extraction of Redis config with all parameters."""
    config_section = {
        "host": "redis.example.com",
        "port": 6380,
        "db": 5,
        "password": "secret123",
        "unix_socket_path": "/var/run/redis.sock",
    }

    result = _extract_redis_config(config_section)

    assert result["host"] == "redis.example.com"
    assert result["port"] == 6380
    assert result["db"] == 5
    assert result["password"] == "secret123"
    assert result["unix_socket_path"] == "/var/run/redis.sock"


def test_extract_redis_config_with_minimal_params():
    """Test extraction with minimal parameters."""
    config_section = {
        "host": "localhost",
        "port": 6379,
        "db": 0,
    }

    result = _extract_redis_config(config_section)

    assert result["host"] == "localhost"
    assert result["port"] == 6379
    assert result["db"] == 0
    assert result["password"] is None
    assert result["unix_socket_path"] is None


def test_extract_redis_config_handles_missing_keys():
    """Test extraction handles missing keys gracefully."""
    config_section = {}

    result = _extract_redis_config(config_section)

    assert result["host"] is None
    assert result["port"] is None
    assert result["db"] is None
    assert result["password"] is None
    assert result["unix_socket_path"] is None


def test_get_redis_params_returns_structured_dict():
    """Test that get_redis_params returns properly structured dict."""
    result = get_redis_params()

    # Should have the three main keys
    assert "indexes" in result
    assert "documents" in result
    assert "use_redis_documents" in result

    # Each should be a dict with connection params
    assert isinstance(result["indexes"], dict)
    assert isinstance(result["documents"], dict)
    assert isinstance(result["use_redis_documents"], bool)

    # Should have standard Redis connection keys
    for key in ["host", "port", "db", "password", "unix_socket_path"]:
        assert key in result["indexes"]
        assert key in result["documents"]


def test_get_redis_params_uses_config_values(config):
    """Test that get_redis_params uses values from config."""
    # Test config should have specific database numbers
    result = get_redis_params()

    # In test environment, databases should be 14 and 15
    assert result["indexes"]["db"] == 14
    assert result["documents"]["db"] == 15


def test_get_redis_params_detects_redis_document_store(config):
    """Test detection of Redis as document store."""
    from addok import ds
    from addok.config import config as addok_config

    result = get_redis_params()

    # Should correctly identify if using Redis for documents
    expected = addok_config.DOCUMENT_STORE == ds.RedisStore
    assert result["use_redis_documents"] == expected



import json

import pytest

from addok.batch import process_documents
from addok.core import search


@pytest.mark.parametrize(
    "input,expected",
    [
        ("rue du 8 mai troyes", False),
        ("8 rue du 8 mai troyes", "8"),
        ("3 rue du 8 mai troyes", "3"),
    ],
)
def test_match_housenumber(input, expected):
    doc = {
        "id": "xxxx",
        "_id": "yyyy",
        "type": "street",
        "name": "rue du 8 Mai",
        "city": "Troyes",
        "lat": "49.32545",
        "lon": "4.2565",
        "housenumbers": {
            "3": {"lat": "48.325451", "lon": "2.25651"},
            "3 bis": {"lat": "48.325451", "lon": "2.25651"},
            "8": {"lat": "48.325451", "lon": "2.25651"},
        },
    }
    process_documents(json.dumps(doc))
    result = search(input)[0]
    assert (result.type == "housenumber") == bool(expected)
    if expected:
        assert result.housenumber == expected



from collections import defaultdict
from itertools import product

from addok.config import config
from addok.db import DB
from addok.helpers import scripts
from addok.pairs import pair_key


def no_tokens_but_housenumbers_and_geohash(helper):
    if not helper.tokens and helper.housenumbers and helper.geohash_key:
        helper.new_bucket([helper.geohash_key], config.BUCKET_MIN)


def no_available_tokens_abort(helper):
    if not helper.tokens:
        return True  # Stop processing.


def only_commons(helper):
    if helper.only_commons:
        # Only common terms, shortcut to search
        keys = [t.db_key for t in helper.tokens]
        if helper.geohash_key:
            keys.append(helper.geohash_key)
            helper.debug("Adding geohash %s", helper.geohash_key)
        if len(keys) == 1 or helper.geohash_key:
            helper.add_to_bucket(keys)
        if helper.bucket_dry and len(keys) > 1:
            # Scan the less frequent token.
            helper.tokens.sort(key=lambda t: t.frequency)
            keys = [t.db_key for t in helper.tokens]
            first = helper.tokens[0]
            if first.frequency < config.INTERSECT_LIMIT:
                helper.debug("Under INTERSECT_LIMIT, force intersect.")
                helper.add_to_bucket(keys)
            else:
                helper.debug("INTERSECT_LIMIT hit, manual scan")
                if helper.filters:
                    # Always consider filters when doing manual intersect.
                    keys = keys + helper.filters
                    # But, hey, can we brute force again?
                    if any(
                        DB.scard(k) < config.INTERSECT_LIMIT for k in helper.filters
                    ):
                        helper.debug("Filters under INTERSECT_LIMIT, force")
                        helper.add_to_bucket(keys)
                        return
                helper.debug('manual scan on "%s"', first)
                ids = scripts.manual_scan(keys=keys, args=[helper.wanted])
                helper.bucket.update(ids)
                helper.debug("%s results after scan", len(helper.bucket))


def bucket_with_meaningful(helper):
    if not helper.meaningful:
        return
    if len(helper.meaningful) == 1 and helper.common and not helper.filters:
        # Avoid running with too less tokens while having commons terms.
        for token in helper.common:
            if token not in helper.meaningful:
                helper.meaningful.append(token)
                break  # We want only one more.
    helper.keys = [t.db_key for t in helper.meaningful]
    if helper.bucket_empty:
        helper.new_bucket(helper.keys, config.BUCKET_MIN)
        if len(helper.bucket) == config.BUCKET_MIN:
            # Do not rerun if bucket with limit 10 has returned less
            # than 10 results.
            helper.new_bucket(helper.keys)
        if (
            not helper.autocomplete
            and helper.has_cream()
            and helper.cream < config.BUCKET_MIN
        ):
            # Do not check cream before computing autocomplete when
            # autocomplete is on.
            # If we have too much cream, do not consider our bucket is good.
            helper.debug("Cream found. Returning.")
            return True
    else:
        helper.add_to_bucket(helper.keys)


def reduce_with_other_commons(helper):
    if helper.only_commons:
        return
    for token in helper.common:  # Already ordered by frequency asc.
        if token not in helper.meaningful and helper.bucket_overflow:
            helper.debug("Now considering also common token %s", token)
            helper.meaningful.append(token)
            helper.keys = [t.db_key for t in helper.meaningful]
            helper.new_bucket(helper.keys)


def ensure_geohash_results_are_included_if_center_is_given(helper):
    if helper.bucket_overflow and helper.geohash_key:
        helper.debug("Bucket overflow and center, force nearby look up")
        helper.add_to_bucket(helper.keys + [helper.geohash_key], max(helper.wanted, 10))


def extend_results_reducing_tokens(helper):
    if helper.bucket_full or helper.has_cream():
        return True
    if not helper.bucket_dry:
        return  # No need.
    # Only if bucket is empty or we have margin on should_match_threshold.
    if (
        helper.bucket_empty
        or len(helper.meaningful) - 1 > helper.should_match_threshold
    ):
        helper.debug("Bucket dry. Trying to remove 1 meaningful token.")

        def sorter(t):
            # First numbers, then by frequency
            return (2 if t.isdigit() else 1, t.frequency)

        helper.meaningful.sort(key=sorter, reverse=True)
        for token in helper.meaningful:
            keys = helper.keys[:]
            keys.remove(token.db_key)
            helper.add_to_bucket(keys)
            if helper.bucket_overflow:
                break

        if helper.bucket_empty and len(helper.meaningful) > 3:
            helper.debug("Bucket still empty, remove 2 meaningful tokens.")
            for token, token2 in product(helper.meaningful, helper.meaningful):
                if token != token2:
                    keys = helper.keys[:]
                    keys.remove(token.db_key)
                    keys.remove(token2.db_key)
                    helper.add_to_bucket(keys)
                    if helper.bucket_overflow:
                        break


def extend_results_extrapoling_relations(helper):
    """Try to extract the bigger group of interlinked tokens.

    Should generally be used at last in the collectors chain.
    """
    if not helper.bucket_dry:
        return  # No need.

    tokens = set(helper.meaningful + helper.common)

    # Sort by average frequency. Needed since we are going to break the loop and we want determinism.
    relations = sorted(
        _extract_manytomany_relations(tokens),
        key=lambda r: sum(t.frequency for t in r) / len(r) if len(r) > 0 else 0
    )

    for relation in relations:
        helper.add_to_bucket([t.db_key for t in relation])
        if helper.bucket_overflow:
            break
    else:
        helper.debug("No relation extrapolated.")


def _extract_manytomany_relations(tokens):
    o2m_relations = _compute_onetomany_relations(tokens)
    m2m_relations = _extrapolate_manytomany_relations(o2m_relations)
    return _deduplicate_sets(m2m_relations)


def _compute_onetomany_relations(tokens):
    relations = defaultdict(list)
    for token in tokens:
        for other in tokens:
            if other == token:
                continue
            if token in relations[other] or DB.sismember(pair_key(token), other):
                relations[token].append(other)
    return relations


def _extrapolate_manytomany_relations(o2m_relations):
    m2m_relations = []
    for origin, others in o2m_relations.items():
        relation = [origin]
        for token in others:
            if all(token in o2m_relations[o] for o in relation):
                relation.append(token)
        # Commons tokens, given that they are "common", are more luckily to
        # create false positives (but we kept them until now because if some
        # other token is not related to some common token, that is a clear true
        # negative).
        relation = set([token for token in relation if not token.is_common])
        if len(relation) > 1:
            m2m_relations.append(relation)
    return m2m_relations


def _deduplicate_sets(sets):
    unique = []
    for set_ in sorted(sets, key=len, reverse=True):
        if not any(set_.issubset(g) for g in unique):
            unique.append(set_)
    return unique



import json

from addok.batch import process_documents, reset
from addok.core import search
from addok.db import DB


def test_process_should_index_by_default(factory):
    doc = factory(skip_index=True, name="Melicocq")
    assert not search("Mélicocq")
    process_documents(json.dumps(doc.copy()))
    assert search("Melicocq")


def test_process_should_deindex_if_action_is_given(factory):
    doc = factory(name="Mélicocq")
    assert search("Mélicoq")
    process_documents(json.dumps({"_action": "delete", "_id": doc["_id"]}))
    assert not search("Mélicoq")


def test_process_should_update_if_action_is_given(factory):
    doc = factory(name="rue de l'avoine")
    assert search("rue")
    doc["_action"] = "update"
    doc["name"] = "avenue de l'avoine"
    process_documents(json.dumps(doc.copy()))
    assert search("avenue")
    assert not search("rue")


def test_reset(factory, monkeypatch):
    class Args:
        force = False

    factory(name="rue de l'avoine")
    assert DB.keys()
    monkeypatch.setitem(__builtins__, "input", lambda *args, **kwargs: "no")
    reset(Args())
    assert DB.keys()
    monkeypatch.setitem(__builtins__, "input", lambda *args, **kwargs: "yes")
    reset(Args())
    assert not DB.keys()


def test_force_reset(factory):
    class Args:
        force = True

    factory(name="rue de l'avoine")
    assert DB.keys()
    reset(Args())
    assert not DB.keys()



import falcon

from addok.config import config, hooks

from .base import CorsMiddleware

config.load()
middlewares = [CorsMiddleware()]
hooks.register_http_middleware(middlewares)
# The name `application` is expected by wsgi by default.
application = api = falcon.App(middleware=middlewares)
# Do not let Falcon split query string on commas.
application.req_options.auto_parse_qs_csv = False
application.req_options.strip_url_path_trailing_slash = True
hooks.register_http_endpoint(api)


def simple(args):
    from wsgiref.simple_server import make_server

    httpd = make_server(args.host, int(args.port), application)
    print("Serving HTTP on {}:{}…".format(args.host, args.port))
    try:
        httpd.serve_forever()
    except (KeyboardInterrupt, EOFError):
        print("Bye!")



from addok.helpers import scripts


def test_manual_scan(factory):
    factory(name="rue de la monnaie", city="Vitry")
    factory(name="La monnaye", city="Saint-Loup-Cammas")
    street1 = factory(name="rue de la monnaie", city="Paris", importance=1)
    street2 = factory(name="rue de la monnaie", city="Condom", importance=0.9)
    results = scripts.manual_scan(keys=["w|monnaie", "w|rue", "w|de"], args=[2])
    assert results == [
        "d|{}".format(street1["_id"]).encode(),
        "d|{}".format(street2["_id"]).encode(),
    ]


def test_manual_scan_with_filter(factory):
    vitry = factory(name="Vitry", type="city")
    factory(name="La monnaye", city="Saint-Loup-Cammas")
    street1 = factory(name="rue de la monnaie", city="Paris", importance=1)
    street2 = factory(name="rue de la monnaie", city="Condom", importance=0.9)
    results = scripts.manual_scan(keys=["w|rue", "w|de", "f|type|street"], args=[2])
    assert results == [
        "d|{}".format(street1["_id"]).encode(),
        "d|{}".format(street2["_id"]).encode(),
    ]
    results = scripts.manual_scan(keys=["w|rue", "w|de", "f|type|whatever"], args=[2])
    assert results == []
    results = scripts.manual_scan(keys=["w|vitry", "f|type|city"], args=[2])
    assert results == ["d|{}".format(vitry["_id"]).encode()]


def test_zinter(factory):
    docs = (
        factory(name="rue de la monnaie", city="Vitry"),
        factory(name="La monnaye", city="Saint-Loup-Cammas"),
        factory(name="rue de la monnaie", city="Paris", importance=1),
        factory(name="rue de la monnaie", city="Condom", importance=0.9),
    )
    results = scripts.zinter(keys=["w|monnaie", "w|rue", "w|de"], args=["tmp", 2])
    assert results == [
        "d|{}".format(docs[2]["_id"]).encode(),
        "d|{}".format(docs[3]["_id"]).encode(),
    ]
    results = scripts.zinter(keys=["w|monnaie", "w|rue", "w|de"], args=["tmp", 3])
    assert results == [
        "d|{}".format(docs[2]["_id"]).encode(),
        "d|{}".format(docs[3]["_id"]).encode(),
        "d|{}".format(docs[0]["_id"]).encode(),
    ]


def test_order_by_frequency(factory):
    factory(name="rue de la monnaie", city="Vitry")
    factory(name="rue des lilas", city="Vitry")
    factory(name="rue des figues", city="Vitry")
    factory(name="rue des lilas", city="Pantin")
    assert scripts.order_by_frequency(
        keys=["w|monnaie", "w|lilas", "w|vitry", "w|rue"]
    ) == [b"w|rue", b"w|vitry", b"w|lilas", b"w|monnaie"]


def test_order_by_max_score(factory):
    factory(name="rue de la monnaie", city="Vitry")
    factory(name="rue des lilas", city="Vitry")
    factory(name="rue des figues", city="Vitry")
    factory(name="rue des lilas", city="Pantin")
    factory(name="Vitry", importance=0.5)
    keys = ["w|monnaie", "w|lilas", "w|vitry", "w|rue"]
    assert scripts.order_by_max_score(keys=keys)[0] == b"w|vitry"



def geojson(result):
    properties = {
        "label": str(result),
    }
    if result._scores:
        properties["score"] = result.score
    for key in result.keys:
        val = getattr(result, key, None)
        if val and key not in ["lat", "lon", "_id"]:
            properties[key] = val
    type_ = result._doc.get("type")
    if type_ and type_ not in properties:
        properties[type_] = properties.get("name")
    housenumber = getattr(result, "housenumber", None)
    if housenumber:
        properties["name"] = "{} {}".format(housenumber, properties.get("name"))
    try:
        properties["distance"] = int(result.distance)
    except ValueError:
        pass
    return {
        "type": "Feature",
        "geometry": {
            "type": "Point",
            "coordinates": [float(result.lon), float(result.lat)],
        },
        "properties": properties,
    }



TOKEN_PREFIX = "w|"


def token_key(s):
    return "{}{}".format(TOKEN_PREFIX, s)


def document_key(s):
    return "d|{}".format(s)


def geohash_key(s):
    return "g|{}".format(s)


def filter_key(k, v):
    return "f|{}|{}".format(k, v)



import os
import importlib
import types
import sys
import warnings


from importlib.metadata import version as get_version, PackageNotFoundError

from addok import hooks, VERSION
from . import default


class Config(dict):

    TESTING = False

    def __init__(self):
        self._post_load_func = []
        self.loaded = False
        self.plugins = [
            "addok.shell",
            "addok.http.base",
            "addok.batch",
            "addok.pairs",
            "addok.fuzzy",
            "addok.autocomplete",
        ]
        super().__init__()
        self.extend_from_object(default)

    def load(self):
        if self.loaded:
            return
        self.loaded = True
        self.load_core_plugins()
        if not Config.TESTING:
            # We don't want to autoload installed plugin during tests.
            hooks.load()  # pragma: no cover
        hooks.preconfigure(self)
        self.load_local()
        self.load_from_env()
        hooks.configure(self)
        self.resolve()
        self.post_process()
        plugins = []
        for name, plugin in hooks.plugins.items():  # pragma: no cover
            if name.startswith("addok."):
                version = VERSION  # Core plugin.
            else:
                try:
                    version = get_version(plugin.__package__)
                except PackageNotFoundError:
                    version = "?"
            plugins.append("{}=={}".format(name, version))
        print("Loaded plugins:\n{}".format(", ".join(plugins)))

    def on_load(self, func):
        self._post_load_func.append(func)
        return func

    def extend_from_object(self, obj):
        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)

    def load_core_plugins(self):
        for path in self.plugins:
            plugin = importlib.import_module(path)
            hooks.register(plugin)

    def load_local(self):
        path = os.environ.get("ADDOK_CONFIG_MODULE") or os.path.join(
            "/etc", "addok", "addok.conf"
        )
        if not os.path.exists(path):
            print('No local config file found in "{}".'.format(path))
            return

        d = types.ModuleType("config")
        d.__file__ = path
        try:
            with open(path) as config_file:
                exec(compile(config_file.read(), path, "exec"), d.__dict__)
        except (IOError, SyntaxError) as e:
            from addok.helpers import red

            print(red("Unable to import {} from " "ADDOK_CONFIG_MODULE".format(path)))
            sys.exit(e)
        else:
            print("Loaded local config from", path)
            self.extend_from_object(d)

    def load_from_env(self):
        for key, value in self.items():
            if key.isupper():
                env_key = "ADDOK_" + key
                typ = type(value)
                if typ in (list, tuple, set):
                    real_type, typ = typ, lambda x: real_type(x.split(","))
                if env_key in os.environ:
                    self[key] = typ(os.environ[env_key])

    def __getattr__(self, key):
        return self.get(key)

    def __setattr__(self, key, value):
        self[key] = value

    def post_process(self):
        # Retrocompat
        if self.SYNONYMS_PATH:
            warnings.warn(
                "SYNONYMS_PATH is deprecated, use SYNONYMS_PATHS instead",
                DeprecationWarning,
            )
            self.SYNONYMS_PATHS.append(self.SYNONYMS_PATH)
        self.FIELDS.extend(self.EXTRA_FIELDS)
        for field in self.FIELDS:
            key = field["key"]
            if field.get("type") == "housenumbers" or key == "housenumbers":
                self.HOUSENUMBERS_FIELD = key
                field["type"] = "housenumbers"
            elif field.get("type") == "name" or key == "name":
                self.NAME_FIELD = key
                field["type"] = "name"
            elif field.get("type") == "id":
                self.ID_FIELD = key
        for func in self._post_load_func:
            func()

    def resolve(self):
        for key in list(self.keys()):
            if key.endswith("_PYPATHS"):
                self.resolve_paths(key)
            elif key.endswith("_PYPATH"):
                self.resolve_path(key)

    def resolve_path(self, key):
        from addok.helpers import import_by_path

        self[key[: -len("_PYPATH")]] = import_by_path(self[key])

    def resolve_paths(self, key):
        from addok.helpers import import_by_path

        self[key[: -len("_PYPATHS")]] = [import_by_path(path) for path in self[key]]


config = Config()



from addok.core import Result, search
from addok.helpers import collectors


def test_should_match_name(street):
    assert not search("Conflans")
    street.update(name="Conflans")
    results = search("Conflans")
    assert results
    result = results[0]
    assert result.name == "Conflans"
    assert result.id == street["id"]


def test_should_match_name_case_insensitive(street):
    assert not search("conflans")
    street.update(name="Conflans")
    assert search("conflans")


def test_should_match_name_with_accent(street):
    assert not search("andrésy")
    street.update(name="Andrésy")
    assert search("andrésy")


def test_should_match_name_without_accent(street):
    assert not search("andresy")
    street.update(name="Andrésy")
    assert search("andresy")


def test_should_give_priority_to_best_match(street, city):
    street.update(name="rue d'Andrésy")
    city.update(name="Andrésy")
    results = search("andresy")
    assert results[0].id == city["id"]


def test_should_give_priority_to_best_match2(street, factory):
    street.update(name="rue d'Andrésy", city="Conflans")
    factory(name="rue de Conflans", city="Andrésy")
    results = search("rue andresy")
    assert len(results) == 2
    assert results[0].id == street["id"]


def test_should_give_priority_to_best_match3(street, factory):
    street.update(name="rue de Lille", city="Douai")
    other = factory(name="rue de Douai", city="Lille")
    results = search("rue de lille douai")
    assert len(results) == 2
    assert results[0].id == street["id"]
    results = search("rue de douai lille")
    assert len(results) == 2
    assert results[0].id == other["id"]


def test_should_be_fuzzy_of_1_by_default(city, config):
    config.FUZZY_KEY_MAP = None
    city.update(name="Andrésy")
    assert search("antresy")
    assert not search("antresu")


def test_fuzzy_should_work_with_inversion(city):
    city.update(name="Andrésy")
    assert search("andreys")


def test_fuzzy_should_match_with_removal(city):
    city.update(name="Andrésy")
    assert search("andressy")


def test_should_give_priority_to_housenumber_if_match(housenumber):
    housenumber.update(name="rue des Berges")
    results = search("rue des berges")
    assert not results[0].housenumber
    results = search("11 rue des berges")
    assert results[0].housenumber == "11"
    assert results[0].type == "housenumber"


def test_return_housenumber_if_number_included_in_bigger_one(factory):
    factory(name="rue 1814", housenumbers={"8": {"lat": "48.3254", "lon": "2.256"}})
    results = search("rue 1814")
    assert not results[0].housenumber
    results = search("8 rue 1814")
    assert results[0].housenumber == "8"


def test_should_do_autocomplete_on_last_term(street):
    street.update(name="rue de Wambrechies", city="Bondues")
    assert search("avenue wambre", autocomplete=True)
    assert not search("wambre avenue", autocomplete=True)


def test_synonyms_should_be_replaced(street, config):
    config.SYNONYMS = {"bd": "boulevard"}
    config.MIN_SCORE = 0
    street.update(name="boulevard des Fleurs")
    assert search("bd")


def test_should_return_results_if_only_common_terms(factory, config):
    config.COMMON_THRESHOLD = 2
    config.INTERSECT_LIMIT = 2
    config.BUCKET_MAX = 3
    street1 = factory(name="rue de la monnaie", city="Vitry")
    street2 = factory(name="rue de la monnaie", city="Paris")
    street3 = factory(name="rue de la monnaie", city="Condom")
    street4 = factory(name="La monnaye", city="Saint-Loup-Cammas")
    results = search("rue de la monnaie")
    ids = [r.id for r in results]
    assert street1["id"] in ids
    assert street2["id"] in ids
    assert street3["id"] in ids
    assert street4["id"] not in ids


def test_should_brute_force_if_common_terms_above_limit(factory, config):
    config.COMMON_THRESHOLD = 2
    config.BUCKET_MAX = 3
    street1 = factory(name="rue de la monnaie", city="Vitry")
    street2 = factory(name="rue de la monnaie", city="Paris")
    street3 = factory(name="rue de la monnaie", city="Condom")
    street4 = factory(name="La monnaye", city="Saint-Loup-Cammas")
    results = search("rue de la monnaie")
    ids = [r.id for r in results]
    assert street1["id"] in ids
    assert street2["id"] in ids
    assert street3["id"] in ids
    assert street4["id"] not in ids


def test_should_use_filter_if_only_common_terms(factory, config):
    config.COMMON_THRESHOLD = 2
    config.INTERSECT_LIMIT = 2
    config.BUCKET_MAX = 3
    street1 = factory(name="rue de la monnaie", city="Vitry")
    street2 = factory(name="rue de la monnaie", city="Paris")
    street3 = factory(name="rue de la monnaie", city="Condom")
    city = factory(name="La monnaie", type="city")
    results = search("la monnaie", type="city")
    ids = [r.id for r in results]
    assert city["id"] in ids
    assert street1["id"] not in ids
    assert street2["id"] not in ids
    assert street3["id"] not in ids


def test_not_found_term_is_autocompleted(factory, config):
    config.COMMON_THRESHOLD = 3
    config.BUCKET_MAX = 3
    factory(name="rue de la monnaie", city="Vitry")
    assert search("rue de la mon")


def test_found_term_is_autocompleted_if_missing_results(factory, config):
    config.COMMON_THRESHOLD = 3
    config.BUCKET_MAX = 3
    factory(name="rue de la montagne", city="Vitry")
    factory(name="rue du mont", city="Vitry")
    assert len(search("rue mont", autocomplete=True)) == 2


def test_found_term_is_not_autocompleted_if_enough_results(factory, config):
    config.COMMON_THRESHOLD = 3
    config.BUCKET_MAX = 3
    montagne = factory(name="rue de la montagne", city="Vitry")
    factory(name="rue du mont", city="Vitry")
    factory(name="rue du mont", city="Paris")
    factory(name="rue du mont", city="Lille")
    results = search("rue mont", limit=2, autocomplete=True)
    ids = [r.id for r in results]
    assert len(ids) == 2
    assert montagne["id"] not in ids


def test_should_autocomplete_if_only_commons_but_geohash(factory, config):
    config.COMMON_THRESHOLD = 3
    config.BUCKET_MAX = 3
    factory(name="rue des tilleuls")
    factory(name="rue des chênes")
    factory(name="rue des hètres")
    factory(name="rue des aulnes")
    factory(name="rue descartes", lon=2.2, lat=48.1)
    results = search("rue des", autocomplete=True, lon=2.2, lat=48.1)
    assert results[0].name == "rue descartes"


def test_should_autocomplete_if_only_housenumbers_but_geohash(factory, config):
    config.COMMON_THRESHOLD = 3
    config.BUCKET_MAX = 3
    config.MIN_SCORE = 0.05
    factory(name="rue des tilleuls", lon=2.256, lat=48.3254)
    factory(name="rue des chênes", lon=2.256, lat=48.3254)
    factory(name="rue des hètres", lon=2.256, lat=48.3254)
    factory(name="rue des aulnes", lon=2.256, lat=48.3254)
    factory(
        name="rue descartes",
        lon=2.256,
        lat=48.3254,
        housenumbers={"11": {"lat": "48.3254", "lon": "2.256"}},
    )
    results = search("11", autocomplete=True, lon=2.256, lat=48.3254)
    assert results[0].name == "rue descartes"


def test_closer_result_should_be_first_for_same_score(factory):
    expected = factory(name="rue de paris", city="Cergy", lat=48.1, lon=2.2)
    factory(name="rue de paris", city="Perpète", lat=-48.1, lon=-2.2)
    factory(name="rue de paris", city="Loin", lat=8.1, lon=42.2)
    results = search("rue de la monnaie", lat=48.1, lon=2.2)
    assert len(results) == 3
    assert results[0].id == expected["id"]


def test_nearby_should_be_included_even_in_overflow(factory, config):
    config.BUCKET_MAX = 3
    config.BUCKET_MIN = 2
    expected = factory(name="Le Bourg", lat=48.1, lon=2.2, importance=0.09)
    factory(name="Le Bourg", lat=-48.1, lon=-2.2, importance=0.1)
    factory(name="Le Bourg", lat=8.1, lon=42.2, importance=0.1)
    factory(name="Le Bourg", lat=10, lon=20, importance=0.1)
    results = search("bourg", lat=48.1, lon=2.2, limit=3, verbose=True)
    assert len(results) == 3
    ids = [r.id for r in results]
    assert expected["id"] in ids


def test_autocomplete_should_give_priority_to_nearby(factory, config):
    config.BUCKET_MAX = 3
    config.BUCKET_MIN = 2
    expected = factory(name="Le Bourg", lat=48.1, lon=2.2, importance=0.09)
    factory(name="Le Bourg", lat=-48.1, lon=-2.2, importance=0.1)
    factory(name="Le Bourg", lat=8.1, lon=42.2, importance=0.1)
    factory(name="Le Bourg", lat=10, lon=20, importance=0.1)
    results = search("bou", lat=48.1, lon=2.2, limit=3, autocomplete=True)
    assert len(results) == 3
    ids = [r.id for r in results]
    assert expected["id"] in ids


def test_document_without_name_should_not_be_indexed(factory):
    doc = factory(skip_index=True, city="Montceau-les-Mines")
    del doc["name"]
    doc.index()
    assert not search("Montceau-les-Mines")


def test_score_is_not_greater_than_one(factory):
    factory(name="rue de paris", importance=1)
    results = search("rue de paris")
    assert len(results) == 1
    assert results[0].score == 1


def test_search_can_be_filtered(factory):
    street = factory(name="rue de Paris", type="street")
    city = factory(name="Paris", type="city")
    results = search("paris", type="street")
    ids = [r.id for r in results]
    assert street["id"] in ids
    assert city["id"] not in ids


def test_search_supports_multi_value_filter(factory):
    street = factory(name="rue de Paris", type="street")
    city = factory(name="Paris", type="city")
    locality = factory(name="Grenelle", type="locality") # Unwanted result
    results = search("paris", type="street+city")
    ids = {r.id for r in results}
    assert street["id"] in ids
    assert city["id"] in ids
    assert locality["id"] not in ids


def test_search_multi_filter_combination_with_other_filters(factory):
    street_75000 = factory(name="rue de Paris", type="street", postcode="75000")
    street_77000 = factory(name="avenue de Paris", type="street", postcode="77000")
    city = factory(name="Paris", type="city", postcode="75000")
    results = search("paris", type="street+city", postcode="75000")
    ids = {r.id for r in results}
    assert street_75000["id"] in ids
    assert city["id"] in ids
    assert street_77000["id"] not in ids


def test_filters_are_stripped(factory):
    street = factory(name="rue de Paris", type="street")
    city = factory(name="Paris", type="city")
    results = search("paris", type="street ")
    ids = [r.id for r in results]
    assert street["id"] in ids


def test_multifilter_with_duplicate_values(factory):
    """Test that duplicate values in multi-filter are deduplicated"""
    street = factory(name="rue de Paris", type="street")
    city = factory(name="Paris", type="city")
    # "street+street+city" should be equivalent to "street+city"
    results = search("paris", type="street+street+city")
    ids = {r.id for r in results}
    assert street["id"] in ids
    assert city["id"] in ids


def test_multifilter_with_spaces(factory):
    """Test that spaces in multi-filter are handled correctly"""
    street = factory(name="rue de Paris", type="street")
    city = factory(name="Paris", type="city")
    locality = factory(name="Grenelle", type="locality")
    # Spaces should be normalized to +
    results = search("paris", type="street + city")
    ids = {r.id for r in results}
    assert street["id"] in ids
    assert city["id"] in ids
    assert locality["id"] not in ids


def test_multifilter_with_empty_values(factory):
    """Test that empty values in multi-filter are ignored"""
    street = factory(name="rue de Paris", type="street")
    city = factory(name="Paris", type="city")
    # "+street++" should be normalized to "street"
    results = search("paris", type="+street++")
    ids = {r.id for r in results}
    assert street["id"] in ids
    assert city["id"] not in ids


def test_multifilter_respects_max_values(factory):
    """Test that multi-filter is limited to MAX_FILTER_VALUES (default: 10)"""
    # Create various types
    street = factory(name="rue de Paris", type="street")
    city = factory(name="Paris", type="city")
    locality = factory(name="Paris Locality", type="locality")
    municipality = factory(name="Paris Municipality", type="municipality")
    
    # Try to use 12 values (default limit is 10), only first 10 should be used
    filter_value = "+".join([
        "street", "city", "locality", "municipality", 
        "zone", "district", "sector", "area",
        "region", "country", "extra1", "extra2"
    ])
    results = search("paris", type=filter_value)
    # Should match at least the first few types created
    ids = {r.id for r in results}
    assert street["id"] in ids
    assert city["id"] in ids


def test_multifilter_case_sensitivity(factory):
    """Test that multi-filter values maintain case"""
    street = factory(name="rue de Paris", type="Street")
    city = factory(name="Paris", type="City")
    locality = factory(name="Paris Locality", type="locality")
    # Filter values should be case-sensitive - "street" won't match "Street"
    results = search("paris", type="Street+City")
    ids = {r.id for r in results}
    assert street["id"] in ids
    assert city["id"] in ids
    # locality should not be in results since it doesn't match the filter
    assert locality["id"] not in ids


def test_housenumber_type_should_enforce_housenumber_match(factory):
    without_housenumber = factory(name="avenue de Paris", type="street")
    with_wrong_housenumber = factory(
        name="boulevard de Paris",
        type="street",
        housenumbers={"12": {"lat": "48.3254", "lon": "2.256"}},
    )
    with_housenumber = factory(
        name="rue de Paris",
        type="street",
        housenumbers={"11": {"lat": "48.3254", "lon": "2.256"}},
    )
    results = search("11 paris", type="housenumber")
    ids = [r.id for r in results]
    assert with_housenumber["id"] in ids
    assert without_housenumber["id"] not in ids
    assert with_wrong_housenumber["id"] not in ids


def test_filter_indexes_multiple_values(factory):
    city = factory(name="Paris", type=["city", "municipality"])
    results = search("paris", type="city")
    ids = [r.id for r in results]
    assert city["id"] in ids
    results = search("paris", type="municipality")
    ids = [r.id for r in results]
    assert city["id"] in ids


def test_housenumber_are_not_computed_if_another_type_is_asked(factory):
    factory(
        name="rue de Bamako",
        type="street",
        housenumbers={"11": {"lat": "48.3254", "lon": "2.256"}},
    )

    results = search("11 rue de bamako")
    assert len(results) == 1
    assert results[0].type == "housenumber"

    results = search("11 rue de bamako", type="housenumber")
    assert len(results) == 1
    assert results[0].type == "housenumber"

    results = search("11 rue de bamako", type="street")
    assert len(results) == 1
    assert results[0].type == "street"


def test_housenumbers_payload_fields_are_exported(config, factory):
    factory(
        name="rue de Paris",
        type="street",
        id="123",
        housenumbers={"1": {"lat": "48.32", "lon": "2.25", "key": "abc"}},
    )
    results = search("rue de paris")
    assert results[0].key == ""
    results = search("1 rue de paris")
    assert results[0].key == "abc"


def test_id_is_overwritten_when_given_in_housenumber_payload(config, factory):
    factory(
        name="rue de Paris",
        type="street",
        id="123",
        housenumbers={"1": {"lat": "48.325", "lon": "2.256", "id": "abc"}},
    )
    results = search("rue de paris")
    assert results[0].id == "123"
    results = search("1 rue de paris")
    assert results[0].id == "abc"


def test_postcode_is_overwritten_when_in_housenumber_payload(config, factory):
    factory(
        name="rue de Paris",
        type="street",
        id="123",
        postcode="12345",
        housenumbers={"1": {"lat": "48.325", "lon": "2.256", "postcode": "54321"}},
    )
    results = search("rue de paris")
    assert results[0].postcode == "12345"
    results = search("1 rue de paris")
    assert results[0].postcode == "54321"


def test_from_id(factory):
    doc = factory(name="avenue de Paris", type="street", id="123")
    result = Result.from_id(doc["_id"])
    assert result.id == "123"


def test_should_compare_with_multiple_values(city, factory):
    city.update(name=["Vernou-la-Celle-sur-Seine", "Vernou"])
    factory(name="Vernou", type="city")
    results = search("vernou")
    assert len(results) == 2
    assert results[0].score == results[1].score


def test_allow_to_set_result_values(factory):
    factory(name="porte des lilas", type="street", id="456")
    results = search("porte des lilas")
    result = results[0]
    result.name = "blah"
    result.score = 22
    # Plugins may need that.
    assert result.name == "blah"
    assert result.score == 22


def test_should_keep_unchanged_name_as_default_label(factory):
    factory(name="Porte des Lilas")
    results = search("porte des lilas")
    str(results[0]) == "Porte des Lilas"


def test_does_not_fail_without_usable_tokens(street):
    assert not search("./.$*")


def test_word_order_priority(factory):
    factory(name="avenue de paris", city="saint-mandé", importance=0.0185)
    factory(name="avenue de saint-mandé", city="paris", importance=0.0463)
    results = search("avenue de paris saint-mandé")
    assert results[0].name == "avenue de paris"

    # Does not work with compare_ngram.
    # Both document have same score:
    # Comparing "avenue de paris saint-mandé france"
    # - with "avenue de saint-mandé paris" => 0.7878787878787878
    # - with "avenue de paris saint-mandé" => 0.7878787878787878
    # results = search("avenue de paris saint-mandé france")
    # assert results[0].name == "avenue de paris"

    results = search("avenue de saint-mandé paris")
    assert results[0].name == "avenue de saint-mandé"


def test_bucket_respects_limit(config, factory):
    # issue #422
    config.BUCKET_MAX = 100
    limit = config.BUCKET_MAX * 2
    fields = {
        "name": "allée des acacias",
        "type": "street",
        "housenumbers": {"1": {"lat": "48.325", "lon": "2.256"}},
    }
    for city in range(0, limit):
        factory(id=str(city), postcode=str(10000 + city), **fields)
    results = search("allée des acacias", limit=limit, autocomplete=True)
    assert len(results) == limit
    results = search("allée des acacias", limit=limit, autocomplete=False)
    assert len(results) == limit


def test_geo_priority(config, factory):
    factory(
        name="Villa Eugène",
        city="Colombes",
        importance=0.0147,
        housenumbers={"13": {"lat": "48.915805", "lon": "2.260938"}},
    )
    factory(
        name="Villa Eugène",
        city="Fontenay",
        importance=0.0191,
        housenumbers={"13": {"lat": "48.879839", "lon": "2.393369"}},
    )
    results = search("13 vla eugène", lat=48.9158, lon=2.2609, autocomplete=True)
    assert results[0].city == "Colombes"
    results = search("13 vla eugène", lat=48.9158, lon=2.2609, autocomplete=False)
    assert results[0].city == "Colombes"


def test_geo_importance_weight(config, factory):
    initial_weight = config.GEO_DISTANCE_WEIGHT
    factory(name="rue descartes", lon=2.2, lat=48.1)
    results = search("rue descartes", lat=48.9158, lon=2.2609)
    initial_geo_score = results[0]._scores["geo_distance"]
    config.GEO_DISTANCE_WEIGHT = 2.0
    results = search("rue descartes", lat=48.9158, lon=2.2609)
    assert (
        results[0]._scores["geo_distance"][0]
        == initial_geo_score[0] * config.GEO_DISTANCE_WEIGHT / initial_weight
    )
    assert results[0]._scores["geo_distance"][1] == config.GEO_DISTANCE_WEIGHT


def test_importance_should_be_minored_if_geohash(factory, config):
    factory(name="rue descartes", lon=2.2, lat=48.1, importance=1)
    results = search("rue descartes")
    assert results[0]._scores["importance"][0] == 0.1
    results = search("rue descartes", lon=2.2, lat=48.1)
    assert results[0]._scores["importance"][0] == 0.010000000000000002


def test_extend_results_reducing_tokens_should_remove_two_tokens(factory, config):
    # Keep the basic bucket_with_meaningful to fill in the various helper
    # properties.
    config.RESULTS_COLLECTORS = [
        collectors.bucket_with_meaningful,
        collectors.extend_results_reducing_tokens,
    ]
    factory(name="quai jules verne", city="saint cyprien")
    factory(name="allee des cyprie", city="larmor plage")
    factory(name="rue jules verne", city="chatelaillon plage")
    factory(name="quai saint truc", city="la plage")
    # "plage" is the bad guy here: it has been seen with every other terms of
    # the search string, but it's not in the searched document.
    results = search("quai jules verne saint cyprie plage")
    assert results[0].name == "quai jules verne"



from addok.http import View


def test_search_without_query_should_return_400(client):
    resp = client.get("/search")
    assert resp.status_code == 400


def test_search_should_return_geojson(client, factory):
    factory(name="rue des avions")
    resp = client.get("/search/", query_string={"q": "avions"})
    assert resp.headers["Content-Type"] == "application/json; charset=utf-8"
    assert resp.json["type"] == "FeatureCollection"
    assert len(resp.json["features"]) == 1
    feature = resp.json["features"][0]
    assert feature["properties"]["name"] == "rue des avions"
    assert feature["properties"]["id"]
    assert feature["properties"]["type"]
    assert feature["properties"]["score"]
    assert "attribution" in resp.json
    assert "licence" in resp.json


def test_search_should_have_cors_headers(client, factory):
    factory(name="rue des avions")
    resp = client.get("/search/", query_string={"q": "avions"})
    assert resp.headers["Access-Control-Allow-Origin"] == "*"
    assert resp.headers["Access-Control-Allow-Headers"] == "X-Requested-With"


def test_search_without_trailing_slash_should_not_be_redirected(client):
    resp = client.get("/search", query_string={"q": "avions"})
    assert resp.status_code == 200


def test_search_can_be_filtered(client, factory):
    factory(name="rue de Paris", type="street")
    factory(name="Paris", type="city")
    resp = client.get("/search/", query_string={"q": "paris", "type": "city"})
    assert resp.json["type"] == "FeatureCollection"
    assert len(resp.json["features"]) == 1
    feature = resp.json["features"][0]
    assert feature["properties"]["name"] == "Paris"
    assert feature["properties"]["type"] == "city"


def test_search_filters_can_be_combined(client, factory):
    factory(name="rue de Paris", type="street", postcode="77000")
    factory(name="avenue de Paris", type="street", postcode="22000")
    factory(name="Paris", type="city")
    resp = client.get(
        "/search/", query_string={"q": "paris", "type": "street", "postcode": "77000"}
    )
    assert resp.json["type"] == "FeatureCollection"
    assert len(resp.json["features"]) == 1
    feature = resp.json["features"][0]
    assert feature["properties"]["name"] == "rue de Paris"
    assert feature["properties"]["type"] == "street"


def test_search_supports_multi_value_filter_via_http(client, factory):
    factory(name="rue de Paris", type="street")
    factory(name="Paris", type="city")
    resp = client.get(
        "/search/", query_string={"q": "paris", "type": "street+city"}
    )
    assert resp.json["type"] == "FeatureCollection"
    assert len(resp.json["features"]) == 2
    types = {feature["properties"]["type"] for feature in resp.json["features"]}
    assert types == {"street", "city"}


def test_centered_search_should_return_center(client, factory):
    factory(name="rue de Paris", type="street")
    resp = client.get("/search/", query_string={"q": "paris", "lat": "44", "lon": "4"})
    assert resp.json["center"] == [4, 44]


def test_reverse_should_return_geojson(client, factory):
    factory(name="rue des avions", lat=44, lon=4)
    resp = client.get("/reverse/", query_string={"lat": "44", "lon": "4"})
    assert resp.json["type"] == "FeatureCollection"
    assert len(resp.json["features"]) == 1
    feature = resp.json["features"][0]
    assert feature["properties"]["name"] == "rue des avions"
    assert feature["properties"]["id"]
    assert feature["properties"]["type"]
    assert feature["properties"]["score"]
    assert "attribution" in resp.json
    assert "licence" in resp.json


def test_reverse_should_also_accept_lng(client, factory):
    factory(name="rue des avions", lat=44, lon=4)
    resp = client.get("/reverse/", query_string={"lat": "44", "lng": "4"})
    assert len(resp.json["features"]) == 1


def test_reverse_without_trailing_slash_should_not_be_redirected(client):
    resp = client.get("/reverse", query_string={"lat": "44", "lon": "4"})
    assert resp.status_code == 200


def test_reverse_can_be_filtered(client, factory):
    factory(lat=48.234545, lon=5.235445, type="street")
    factory(lat=48.234546, lon=5.235446, type="city")
    resp = client.get(
        "/reverse/",
        query_string={"lat": "48.234545", "lon": "5.235446", "type": "city"},
    )
    assert resp.json["type"] == "FeatureCollection"
    assert len(resp.json["features"]) == 1
    feature = resp.json["features"][0]
    assert feature["properties"]["type"] == "city"


def test_reverse_supports_multi_value_filter(client, factory):
    """Test that reverse geocoding supports multi-value filters with OR logic"""
    street = factory(name="rue des avions", lat=48.2345, lon=5.2354, type="street")
    city = factory(name="Ville", lat=48.2345, lon=5.2354, type="city")
    locality = factory(name="Lieu-dit", lat=48.2345, lon=5.2354, type="locality")
    resp = client.get(
        "/reverse/",
        query_string={"lat": "48.2345", "lon": "5.2354", "type": "street+city", "limit": "10"},
    )
    assert resp.json["type"] == "FeatureCollection"
    assert len(resp.json["features"]) == 2
    types = {feature["properties"]["type"] for feature in resp.json["features"]}
    assert types == {"street", "city"}


def test_reverse_multi_filter_combination(client, factory):
    """Test that reverse geocoding combines multiple filters with AND logic"""
    factory(name="rue A", lat=48.2345, lon=5.2354, type="street", postcode="75001")
    factory(name="rue B", lat=48.2345, lon=5.2354, type="street", postcode="75002")
    factory(name="Ville C", lat=48.2345, lon=5.2354, type="city", postcode="75001")
    resp = client.get(
        "/reverse/",
        query_string={
            "lat": "48.2345",
            "lon": "5.2354",
            "type": "street+city",
            "postcode": "75001",
            "limit": "10",
        },
    )
    assert resp.json["type"] == "FeatureCollection"
    # Should only match: rue A (street+75001) and Ville C (city+75001)
    # Should NOT match: rue B (street but wrong postcode)
    assert len(resp.json["features"]) == 2
    names = {feature["properties"]["name"] for feature in resp.json["features"]}
    assert names == {"rue A", "Ville C"}


def test_reverse_should_have_cors_headers(client, factory):
    factory(name="rue des avions", lat=44, lon=4)
    resp = client.get("/reverse/", query_string={"lat": "44", "lng": "4"})
    assert resp.headers["Access-Control-Allow-Origin"] == "*"
    assert resp.headers["Access-Control-Allow-Headers"] == "X-Requested-With"


def test_reverse_without_lat_or_lng_should_return_400(client, factory):
    resp = client.get("/reverse/", query_string={"lat": "44"})
    assert resp.status_code == 400
    resp = client.get("/reverse/", query_string={"lng": "4"})
    assert resp.status_code == 400


def test_view_should_expose_config(config):
    config.NEW_PROPERTY = "ok"
    assert View.config.NEW_PROPERTY == "ok"


def test_geojson_should_return_housenumber_payload(client, factory, config):
    factory(
        name="rue de Paris",
        type="street",
        id="123",
        housenumbers={"1": {"lat": "48.32", "lon": "2.25", "key": "abc"}},
    )
    resp = client.get("/search/", query_string={"q": "rue de paris"})
    assert "key" not in resp.json["features"][0]["properties"]
    resp = client.get("/search/", query_string={"q": "1 rue de paris"})
    assert resp.json["features"][0]["properties"]["key"] == "abc"


def test_geojson_should_keep_housenumber_parent_name(client, factory):
    factory(
        name="rue de Paris",
        type="street",
        id="123",
        housenumbers={"1": {"lat": "48.32", "lon": "2.25"}},
    )
    factory(
        name="Le Vieux-Chêne",
        type="locality",
        id="124",
        housenumbers={"2": {"lat": "48.22", "lon": "2.22"}},
    )
    resp = client.get("/search/", query_string={"q": "1 rue de paris"})
    assert resp.json["features"][0]["properties"]["name"] == "1 rue de Paris"
    assert resp.json["features"][0]["properties"]["street"] == "rue de Paris"
    resp = client.get("/search/", query_string={"q": "2 Le Vieux-Chêne"})
    props = resp.json["features"][0]["properties"]
    assert props["name"] == "2 Le Vieux-Chêne"
    assert props["locality"] == "Le Vieux-Chêne"


def test_search_should_not_split_querystring_on_commas(client, factory):
    factory(name="rue des avions", housenumbers={"18": {"lat": "48.22", "lon": "2.22"}})
    # Pass query string as a string to avoid additional encoding by the test client.
    resp = client.get("/search/", query_string="q=18, rue des avions")
    props = resp.json["features"][0]["properties"]
    assert props["label"] == "18 rue des avions"
    assert resp.json["query"] == "18, rue des avions"


def test_query_string_length_should_be_checked(client, config):
    config.QUERY_MAX_LENGTH = 10
    resp = client.get("/search/", query_string={"q": "this is too long"})
    assert resp.status_code == 413
    assert resp.json["title"] == "Query too long, 16 chars, limit is 10"


def test_geojson_should_have_document_type_as_key(client, factory):
    factory(
        name="rue de Paris",
        type="street",
        id="123",
        housenumbers={"1": {"lat": "48.32", "lon": "2.25"}},
    )
    factory(name="rue de foobar", type="foo", id="123", foo="bar")
    resp = client.get("/search/", query_string={"q": "1 rue de paris"})
    properties = resp.json["features"][0]["properties"]
    assert properties["name"] == "1 rue de Paris"
    assert properties["street"] == "rue de Paris"
    resp = client.get("/search/", query_string={"q": "rue de paris"})
    properties = resp.json["features"][0]["properties"]
    assert properties["name"] == "rue de Paris"
    assert properties["street"] == "rue de Paris"
    resp = client.get("/search/", query_string={"q": "rue de foobar"})
    properties = resp.json["features"][0]["properties"]
    assert properties["name"] == "rue de foobar"
    assert properties["foo"] == "bar"  # Not overridden.


def test_search_should_catch_invalid_lat(client):
    resp = client.get("/search?q=blah&lat=invalid&lon=3.21")
    assert resp.status_code == 400
    assert resp.json == {
        "description": 'The "lat" parameter is invalid. invalid value',
        "title": "Invalid parameter",
    }


def test_search_should_catch_invalid_lon(client):
    resp = client.get("/search?q=blah&longitude=invalid&lat=3.21")
    assert resp.status_code == 400
    assert resp.json == {
        "description": 'The "longitude" parameter is invalid. invalid value',
        "title": "Invalid parameter",
    }


def test_search_should_catch_out_of_range_lon(client):
    resp = client.get("/search?q=blah&longitude=200&lat=3.21")
    assert resp.status_code == 400
    assert resp.json == {
        "description": 'The "lon" parameter is invalid. out of range',
        "title": "Invalid parameter",
    }

def test_health_should_return_ok(client):
    resp = client.get("/health")
    assert resp.status_code == 200
    assert resp.json['status'] == "HEALTHY"
